// CamImage.cpp

#include "CamImage.h"
#include "JPEGLoader.h"
#include "EdgeFinder.h"

#include <cstdlib>



// Construct
CamImage::CamImage(int imageId, char *data, int len) :
	id(imageId),
	
	imageData(reinterpret_cast<unsigned char*>(data)),
	dataLen(len),
	
	pixels(0),
	w(0),
	h(0),
	
	bmp(0),
	bmpLen(0),
	
	m_usage(1),
	
	m_decoded(false),
	m_decoding(false),
	
	m_hasBmp(false),
	m_makingBmp(false)
{
}// CamImage::CamImage


// Destruct
CamImage::~CamImage() {
	delete[] imageData;
	delete[] pixels;
	delete[] bmp;
}// CamImage::~CamImage


// Acquire reference to the image
CamImage* CamImage::acquire() {
	m_mutex.lock();
	m_usage++;
	m_mutex.unlock();
	
	return this;
}// CamImage::acquire


// Release reference (delete when count goes to zero)
void CamImage::release() {
	bool delNow = false;
	
	m_mutex.lock();
	m_usage--;
	if(m_usage == 0)
		delNow = true;
	m_mutex.unlock();
	
	if(delNow) {
		delete this;
	}
}// CamImage::release


// Decode image data from JPEG
// Returns true if the image pixels area ready for access
// If a thread requests decodeImage() while another thread is working on decoding,
// this method will return false
// If decoding has not yet been started this method will perform a blocking
// decode on the calling thread
bool CamImage::decodeImage() {
	bool	done;
	bool	decodeNow = false;
	
	// Allow only one thread at a time to decode
	m_mutex.lock();
	done = m_decoded;
	if(!done) {
		if(!m_decoding) {
			m_decoding = true;
			decodeNow = true;
		}
	}
	m_mutex.unlock();
	
	if(done)
		return true;
	
	// Decode in this thread
	if(decodeNow) {
		int temp;
		
		if(loadJPEG(imageData, dataLen, &pixels, &temp, &w, &h)) {
			m_mutex.lock();
			m_decoded = true;
			done = true;
			m_mutex.unlock();
		}
		
		m_mutex.lock();
		m_decoding = false;
		m_mutex.unlock();
	}
	
	return done;
}// CamImage::decodeImage


// Make BMP image from image data
// Thread-safe like the decodeImage() method
bool CamImage::makeBmp() {
	bool	done;
	bool	makeNow = false;
	
	m_mutex.lock();
	done = m_decoded;
	m_mutex.unlock();
	
	if(!done)
		return false;
	
	m_mutex.lock();
	done = m_hasBmp;
	if(!done) {
		if(!m_makingBmp) {
			m_makingBmp = true;
			makeNow = true;
		}
	}
	m_mutex.unlock();
	
	// Create BMP data in this thread
	if(makeNow) {
		bmpLen = 26+h*w*3;
		
		bmp = new char[bmpLen];
		
		bmp[0] = 0x42;
		bmp[1] = 0x4d;
		
		*(unsigned int*)(bmp+2) = 26+h*w*3;
		*(unsigned int*)(bmp+10) = 26;
		*(unsigned int*)(bmp+14) = 12;
		*(unsigned short*)(bmp+18) = w;
		*(unsigned short*)(bmp+20) = h;
		*(unsigned short*)(bmp+22) = 1;
		*(unsigned short*)(bmp+24) = 24;
		
		char *idata = bmp+26;
		for(int i=0;i<h;i++) {
			for(int j=0;j<w;j++) {
				*idata++ = *(char*)(pixels + (h-1-i)*w*3 + j*3 + 0);
				*idata++ = *(char*)(pixels + (h-1-i)*w*3 + j*3 + 1);
				*idata++ = *(char*)(pixels + (h-1-i)*w*3 + j*3 + 2);
			}
		}
		
		// Draw a grid on the image
		idata = bmp+26;
		for(int i=1;i<20;i++) {
			int y = (i * h) / 20;
			for(int j=0;j<w;j++) {
				*(unsigned char*)(idata + y*w*3 + j*3 + 0) = 0xff;
				*(unsigned char*)(idata + y*w*3 + j*3 + 1) = 0xff;
				*(unsigned char*)(idata + y*w*3 + j*3 + 2) = 0xff;
			}
		}
		for(int i=1;i<6;i++) {
			int x = (i * w) / 6;
			for(int j=0;j<h;j++) {
				*(unsigned char*)(idata + j*w*3 + x*3 + 0) = 0xff;
				*(unsigned char*)(idata + j*w*3 + x*3 + 1) = 0xff;
				*(unsigned char*)(idata + j*w*3 + x*3 + 2) = 0xff;
			}
		}
		
		done = true;
		
		m_mutex.lock();
		m_hasBmp = true;
		m_makingBmp = false;
		m_mutex.unlock();
	}
	
	return done;
}// CamImage::decodeImage
