// EdgeFinder.cpp

// Finds edges in an image, and calculates possible humans in the image, by
// searching for paralell vertical edges that look like human legs

#include "EdgeFinder.h"
#include "vec2.h"
#include <stdio.h>
#include <cmath>


// Get world distances from an image pixel row
bool getDists(double pixy, double imgy, vec2d &dists) {
	// This table is manually constructed from measurements
	// for a particular camera-position on the robot
	// (it must be calibrated if the camera is moved)
	static const int th = 8;
	static const vec2d table[8] = {
		vec2d(0,0),
		vec2d(0,0),
		vec2d(0.03, 0.1),
		vec2d(0.04, 0.12),
		vec2d(0.05, 0.14),
		vec2d(0.06, 0.2),
		vec2d(0.07, 0.22),
		vec2d(0.08, 0.25)
	};
	
	// Linearly interpolate to find an estimate of the distance
	// to the point on the floor at the row pixy
	double y = (pixy * static_cast<double>(th)) / imgy;
	
	int iy = static_cast<int>(y);
	if(iy < 0) iy = 0;
	if(iy > static_cast<int>(imgy)-2) iy = static_cast<int>(imgy)-2;
	
	y -= std::floor(y);
	
	vec2d b0 = table[iy];
	vec2d b1 = table[iy+1];
	
	if(b0.x == 0.0 || b1.x == 0.0) return false;
	
	dists = b0 + (b1 - b0) * y;
	
	return true;
}// getDists


// Find targets from vertical edges in image data
// Output to 'redges'
// Targets are pairs of vertical edges at appropriate distances to be probable human legs
void findEdges(unsigned char *data, int w, int h, std::vector<Edge> &redges, EdgeSettings *settings) {
	std::vector<Edge> edges;
	std::vector<vec2i> targets;
	
	// Add color channels
	int *gray = new int[h*w];
	int *edgepoints = new int[h*w];
	for(int i=0;i<h*w;i++) {
		int bb = static_cast<int>(data[i*3+0]);
		int gg = static_cast<int>(data[i*3+1]);
		int rr = static_cast<int>(data[i*3+2]);
		int gr = (rr + gg + bb);
		gray[i] = gr;
	}
	
	// Gaussian 5x5 blur (remove noise)
	static const int b[25] = {2, 4, 5, 4, 2, 4, 9, 12, 9, 4, 5, 12, 15, 12, 5, 4, 9, 12, 9, 4, 2, 4, 5, 4, 2};
	int *temp = new int[h*w];
	for(int i=2;i<h-2;i++) {
		for(int j=2;j<w-2;j++) {
			int val = 0;
			int mi = 0;
			for(int y=i-2;y<=i+2;y++) {
				for(int x=j-2;x<=j+2;x++) {
					val += gray[y*w+x] * b[mi++];
				}
			}
			temp[i*w+j] = val / 159;
		}
	}
	delete[] gray;
	gray = temp;
	
	// Max contrast
	int maxpix = 0;
	int minpix = 1000;
	for(int i=2;i<h-2;i++) {
		temp = (gray + i*w);
		for(int j=2;j<w-2;j++) {
			int pix = temp[j];
			
			if(pix < minpix) minpix = pix;
			if(pix > maxpix) maxpix = pix;
		}
	}
	int difval = maxpix - minpix;
	for(int i=0;i<h;i++) {
		temp = (gray + i*w);
		for(int j=0;j<w;j++) {
			int t = temp[j] - minpix;
			t *= 765;
			t /= difval;
			temp[j] = t;
		}
	}
	
	// Find edge pixels
	int thres = settings->thres;
	
	for(int i=2;i<h-4;i++) {
		for(int j=3;j<w-4;j++) {
			int pix = gray[i*w+j];
			
			int dif = std::abs(pix - ((gray[i*w+j+1] + gray[i*w+j+2])>>1));
			
			if(dif >= thres)
				edgepoints[i*w+j] = dif;
			else
				edgepoints[i*w+j] = 0;
		}
	}
	
	// Used to skip pixels that are already found to belong to an edge
	// if the same pixels are considered later in the algorithm
	int *adds = new int[h*w];
	for(int i=0;i<h*w;i++)
		adds[i] = 1;
	
	// Find vertical edges
	int minEdgeLen = settings->minEdgeLen;
	int minStartLen = settings->minStartLen;
	int maxYMiss = settings->maxYMiss;
	int minCenterMoveY = settings->minCenterMoveY;
	
	// Loop over entire image
	for(int i=2;i<h-4;i++) {
		for(int j=3;j<w-4;j++) {
			// Edge pixel found
			if(edgepoints[i*w+j]) {
				Edge e;
				
				// Edge starts as one pixel
				e.sx = j;
				e.sy = i;
				e.ex = j;
				e.ey = i+1;
				
				int cx = j;// Current x
				int ocx = cx;// Current edge-middle x
				int cy = i;// Y at last update of ocx
				int cl = ocx-1;// Left-most accepted edge-pixel to continue edge
				int cr = ocx+1;// Right-most
				
				int ylen = 0;
				int ymiss = 0;
				int missx, missy;
				
				// Scan downwards
				for(int y=i+1;y<h-4;y++) {
					bool end = true;
					int cthres = thres;
					
					// Scan this line in the image close to the vertical edge
					for(int x=cx-2;x<=cx+2;x++) {
						// Skip if in proximity of an already found edge
						if(adds[y*w+x] > 1) break;
						
						// Check for edge pixel (intensity takes priority over order)
						if(edgepoints[y*w+x] >= cthres) {
							cthres = gray[y*w+x] + 1;
							
							// Update position
							cx = x;
							
							// Check if edge is bending from vertical
							if(cx != ocx) {
								// Check if to allow edge to move further horizontally
								// (should be a reasonably straight vertical line, but allow some offset)
								if(y-cy > minCenterMoveY) {
									// Update to new edge center
									// (so that we can find somewhat diagonal lines)
									ocx = cx;
									cy = y;
									cl = ocx-1;
									cr = ocx+1;
								}
							}
							
							end = false;
							
							adds[y*w+x] = 7;
						}
					}
					
					// If edge doesn't continue
					if(end) {
						if(ymiss == 0) {
							missx = cl+1;
							missy = y;
						}
						
						ymiss++;
						
						// Allow some gaps in the edge, if it's already long enough
						if(ylen < minStartLen || ymiss > maxYMiss) {
							// Check if edge is long enough to save
							if(missy > i+minEdgeLen) {
								// Update edge end-point
								e.ex = missx;
								e.ey = missy;
								edges.push_back(e);
							}
							
							break;
						}
					}
					else {
						if(ymiss > 0) {
							ymiss = 0;
							//ylen = 0;
						}
						ylen++;
					}
				}
			}
		}
	}
	
	// Find targets from edges
	int minXDist = settings->minXDist;
	int maxXDist = settings->maxXDist;
	int dirNum = settings->dirNum;
	int dirDen = settings->dirDen;
	
	// Loop over all edges, checking for pairs (N^2)
	for(size_t i=0;i<edges.size();++i) {
		const Edge edge(edges[i]);
		
		for(size_t j=i+1;j<edges.size();++j) {
			const Edge edge2(edges[j]);
			const Edge *e0, *e1;
			
			// Sort pair from left to right
			if(edge2.ex < edge.ex) {
				e0 = &edge2;
				e1 = &edge;
			}
			else {
				e0 = &edge;
				e1 = &edge2;
			}
			
			// Get the distance from the robot at the row in the image where
			// these edges end (used to estimate real-world distance between edges)
			vec2d dists;
			if(!getDists((e0->ey+e1->ey)/2.0, h, dists))
				continue;
			
			// Calculate the range of the distance between the edge
			// that we regard as a probable human leg
			minXDist = static_cast<int>(dists.x * w);
			maxXDist = static_cast<int>(dists.y * w);
			
			if(
				e1->ex > e0->ex+minXDist &&
				e1->ex < e0->ex+maxXDist &&
				e1->ey > e0->ey-((e1->ex-e0->ex)*dirNum)/dirDen &&
				e1->ey < e0->ey+((e1->ex-e0->ex)*dirNum)/dirDen
			) {
				// Skip very light highlights, so that lamp-reflections
				// are not regarded as a target (estimate, can be calibrated)
				int sx = e0->ex + 10;
				int sy = e0->ey - 15;
				bool go = true;
				if(sx < w && sy >= 0) {
					if(gray[sy*w+sx] > 700)
						go = false;
				}
				sx = e0->ex - 10;
				if(sx >= 0 && sy >= 0) {
					if(gray[sy*w+sx] > 700)
						go = false;
				}
				
				sx = e1->ex + 10;
				sy = e1->ey - 15;
				if(sx < w && sy >= 0) {
					if(gray[sy*w+sx] > 700)
						go = false;
				}
				sx = e1->ex - 10;
				if(sx >= 0 && sy >= 0) {
					if(gray[sy*w+sx] > 700)
						go = false;
				}
				
				if(go) {
					Edge e;
					
					e.ex = (e0->ex+e1->ex)>>1;
					e.ey = (e0->ey+e1->ey)>>1;
					
					redges.push_back(e);
				}
			}
		}
	}
	
	delete[] adds;
	delete[] gray;
	delete[] edgepoints;
}// findEdges
