// HTTPConnection.cpp

// The HTTP connection handles an admin-connection via HTTP
// It parses the request and sends a response, which can be
// an image of the map etc.
// The request can also be an admin-order for the robot to
// move to a certain position etc, in which case the HTTPConnection
// passes this order on to the RobotHandler
// HTTPConnection accepts connections passed to it through
// 'handleConnection', and then handles that socket in 'handleNetwork',
// which is called each update-cycle
// It was previously asynchronous but for performance-reasons it is now
// not (locks are expensive, and nothing to gain unless there are multiple cores)

#include "HTTPConnection.h"
#include "CamImage.h"
#include "ImageProvider.h"
#include "ProbabilityMap.h"
#include "RobotHandler.h"

void doShutdown();



// Construct
HTTPConnection::HTTPConnection(
	ProbabilityMap *map,
	RobotHandler *robotHandler,
	ImageProvider *imgProvider
) :
	m_map(map),
	m_robotHandler(robotHandler),
	m_imgProvider(imgProvider),
	
	m_exit(false),
	m_hasSock(false),
	m_sock(0),
	
	m_maxBytes(2048),
	m_data(0),
	m_nBytes(0)
{
	m_mapData = new char[1024+m_map->height()*m_map->width()*3];
	
	if(m_maxBytes > 0)
		m_data = new char[m_maxBytes];
}// HTTPConnection::HTTPConnection


// Destruct
HTTPConnection::~HTTPConnection() {
	delete[] m_mapData;
	
	delete[] m_data;
	
	for(size_t i=0;i<m_sockQueue.size();++i) {
		m_sockQueue[i]->close();
		delete m_sockQueue[i];
	}
}// HTTPConnection::~HTTPConnection


// Handle connection
bool HTTPConnection::handleConnection(ArSocket *sock) {
	bool success = false;
	
	m_mutex.lock();
	
	if(!m_hasSock && !m_exit) {
		m_sock = sock;
		m_hasSock = true;
		
		m_nBytes = 0;
		
		success = true;
	}
	else {
		m_sockQueue.push_back(sock);
		
		success = true;
	}
	
	m_mutex.unlock();
	
	return success;
}// HTTPConnection::handleConnection


// Close connection (must only be called on same thread as handleNetwork)
void HTTPConnection::closeNow() {
	m_mutex.lock();
	if(m_hasSock) {
		m_sock->close();
		delete m_sock;
		m_hasSock = false;
	}
	m_exit = true;
	m_mutex.unlock();
}// HTTPConnection::closeNow



// Handle network
void HTTPConnection::handleNetwork() {
	bool go = false;
	
	m_mutex.lock();
	if(m_hasSock)
		go = true;
	m_mutex.unlock();
	
	if(go) {
		bool bail = false;
		
		do {
			// Handle incoming data
			bail = handleRead();
			if(bail) break;
		} while(false);
		
		// On error, or disconnect, close socket
		if(bail) {
			m_mutex.lock();
			m_sock->close();
			delete m_sock;
			m_hasSock = false;
			
			if(m_sockQueue.size() > 0) {
				m_sock = m_sockQueue[m_sockQueue.size()-1];// Lifo.. should be fifo
				m_sockQueue.pop_back();
				m_hasSock = true;
				
				m_nBytes = 0;
			}
			m_mutex.unlock();
		}
	}
}// HTTPConnection::handleNetwork


// Handle read
bool HTTPConnection::handleRead() {
	bool bail = false;
	
	// Read and handle incoming data
	int readCount = m_sock->read(m_data+m_nBytes, m_maxBytes-m_nBytes, 1);
	
	while(readCount > 0) {
		m_nBytes += readCount;
		
		// Try to parse an HTTP GET request
		int result = parseGet(m_data, m_nBytes);
		
		// If an HTTP request was successfully parsed, respond and close connection
		if(result > 0) {
			sendFile();
			bail = true;
			break;
		}
		
		if(m_nBytes < m_maxBytes)
			readCount = m_sock->read(m_data+m_nBytes, m_maxBytes-m_nBytes, 1);
		else {
			bail = true;
			break;
		}
	}
	
	if(readCount < 0) bail = true;
	
	return bail;
}// HTTPConnection::handleRead


// Parse GET request
int HTTPConnection::parseGet(const char *data, int len) {
	for(int i=3;i<len;i++) {
		// \r\n\r\n signals end of header
		if(data[i-3] == '\r' && data[i-2] == '\n' && data[i-1] == '\r' && data[i] == '\n') {
			return parseHeader(data, i);
		}
	}
	
	return 0;
}// HTTPConnection::parseGet



// Parse header

int HTTPConnection::parseHeader(const char *data, int len) {
	// Parse each line in the header
	for(int i=1;i<len;i++) {
		if(data[i-1] == '\r' && data[i] == '\n') {
			return parseLine(data, i-2);
		}
	}
	
	return -1;
}// HTTPConnection::parseHeader



// Parse line

int HTTPConnection::parseLine(const char *data, int len) {
	// Parse GET-line and save the requested file-name
	if(data[0] == 'G' && data[1] == 'E' && data[2] == 'T' && data[3] == ' ' && data[4] == '/') {
		for(int i=4;i<len&&i<250;i++) {
			if(data[i] == ' ') {
				m_requestFile = std::string(data+4, i-4);
				
				return 1;
			}
		}
	}
	
	return -1;
}// HTTPConnection::parseLine



// Send the file requested
// This function is bloated.. should keep responses as files or something
// It's just a long stream of if-statements comparing the requested file
// to the handled request, and a response is sent depending on what request
// it is
void HTTPConnection::sendFile() {
	//char *data;
	int len;
	const char *movetoFile = "/moveto.bin";
	size_t movetoLen = strlen(movetoFile);
	const char *orderFile = "/order.bin";
	size_t orderLen = strlen(orderFile);
	const char *camFile = "/setcam.bin";
	size_t camLen = strlen(camFile);
	const char *sonarFile = "/setsonar.bin";
	size_t sonarLen = strlen(sonarFile);
	
	// The index-page
	// This string contains a bunch of java-script and AJAX to dynamically handle
	// updating the map etc, and allow for clicking in the map-image to tell the robot
	// to move. It is browser-dependent.. currently works in IE as it interprets clicks
	// as relative pixel-positions to the image top/left
	// Firefox does something else... a dynamic browser-check at the least is probably necessary to achieve
	// cross-platformness
	if(m_requestFile.compare(0, m_requestFile.length(), "/", 1) == 0) {
		//const char *content = "<html><head><title>Robot Admin</title></head><body><h1>Robot Admin</h1><img src=\"map.bmp\" border=1></body></html>";
		const char *content =
"<html>\r\n"
"<head>\r\n"
"<title>Robot Admin</title>\r\n"
"<script language=\"JavaScript\">\r\n"
"function getCoords(e) {\r\n"
"	var map = document.getElementById(\"mapimage\");\r\n"
"	var x = e.offsetX ? e.offsetX : (e.pageX - map.offsetLeft);\r\n"
"	var y = e.offsetY ? (map.height - e.offsetY) : (-e.pageY + map.offsetTop + map.height);\r\n"
"	var xmlhttp = new XMLHttpRequest();\r\n"
"	var str = \"/moveto.bin?x=\" + x + \";y=\" + y;\r\n"
"	xmlhttp.open(\"GET\", str);\r\n"
"	xmlhttp.onreadystatechange=function() {\r\n"
"		if(xmlhttp.readyState==4) {\r\n"
"			//alert(xmlhttp.responseText);\r\n"
"		}\r\n"
"	}\r\n"
"	xmlhttp.send(null);\r\n"
"}\r\n"
"function sendOrders(orderVal) {\r\n"
"	var xmlhttp = new XMLHttpRequest();\r\n"
"	var str = \"/order.bin?o=\" + orderVal;\r\n"
"	xmlhttp.open(\"GET\", str);\r\n"
"	xmlhttp.onreadystatechange=function() {\r\n"
"		if(xmlhttp.readyState==4) {\r\n"
"			//alert(xmlhttp.responseText);\r\n"
"		}\r\n"
"	}\r\n"
"	xmlhttp.send(null);\r\n"
"}\r\n"
"function sendRequest(reqStr) {\r\n"
"	var xmlhttp = new XMLHttpRequest();\r\n"
"	xmlhttp.open(\"GET\", reqStr);\r\n"
"	xmlhttp.onreadystatechange=function() {\r\n"
"		if(xmlhttp.readyState==4) {\r\n"
"			//alert(xmlhttp.responseText);\r\n"
"		}\r\n"
"	}\r\n"
"	xmlhttp.send(null);\r\n"
"}\r\n"
"function loadRobotPos() {\r\n"
"	var map = document.getElementById(\"mapimage\");\r\n"
"	var rpos = document.getElementById(\"robotpos\");\r\n"
"	var rhead = document.getElementById(\"robothead\");\r\n"
"	var xmlhttp = new XMLHttpRequest();\r\n"
"	var getstr = \"/getcoords.bin\"\r\n"
"	xmlhttp.open(\"GET\", getstr);\r\n"
"	xmlhttp.onreadystatechange=function() {\r\n"
"		if(xmlhttp.readyState==4) {\r\n"
"			var str = xmlhttp.responseText;\r\n"
"			var pos = str.split(\",\");\r\n"
"			var xx = parseInt(pos[0]);\r\n"
"			var yy = map.height - parseInt(pos[1]);\r\n"
"			rpos.style.left = xx - 5;\r\n"
"			rpos.style.top = yy - 5;\r\n"
"			var rang = parseInt(pos[2]);\r\n"
"			rhead.style.left = xx - 3 + 5 * Math.cos(rang * 0.001);\r\n"
"			rhead.style.top = yy - 3 - 5 * Math.sin(rang * 0.001);\r\n"
"		}\r\n"
"	}\r\n"
"	xmlhttp.send(null);\r\n"
"	setTimeout('loadRobotPos()', 2000);\r\n"
"}\r\n"
"function loadRobotStatus() {\r\n"
"	var sonarstat = document.getElementById(\"sonarstatus\");\r\n"
"	var camstat = document.getElementById(\"camstatus\");\r\n"
"	var rorders = document.getElementById(\"robotorders\");\r\n"
"	var raction = document.getElementById(\"robotaction\");\r\n"
"	var xmlhttp = new XMLHttpRequest();\r\n"
"	var getstr = \"/getstatus.bin\"\r\n"
"	xmlhttp.open(\"GET\", getstr);\r\n"
"	xmlhttp.onreadystatechange=function() {\r\n"
"		if(xmlhttp.readyState==4) {\r\n"
"			var str = xmlhttp.responseText;\r\n"
"			var stat = str.split(\",\");\r\n"
"			sonarstat.innerHTML = stat[0];\r\n"
"			camstat.innerHTML = stat[1];\r\n"
"			rorders.innerHTML = stat[2];\r\n"
"			raction.innerHTML = stat[3];\r\n"
"		}\r\n"
"	}\r\n"
"	xmlhttp.send(null);\r\n"
"	setTimeout('loadRobotStatus()', 2000);\r\n"
"}\r\n"
"function reloadImgs() {\r\n"
"	document.getElementById(\"mapimage\").src = \"map.bmp?\" + (new Date()).getTime();\r\n"
//"	document.getElementById(\"camimage\").src = \"camimage.jpg?\" + (new Date()).getTime();\r\n"
"	setTimeout('reloadImgs()', 2000);\r\n"
"}\r\n"
"</script>\r\n"
"</head>\r\n"
"<body onLoad=\"loadRobotPos();loadRobotStatus();reloadImgs()\">\r\n"
"<h1>Robot Admin <input type=\"submit\" value=\"Shutdown\" onClick=\"window.location='shutdown.bin'\"></h1>\r\n"
"<table border=0 cellspacing=0 cellpadding=0><tr><td width=200>\r\n"
"<b>Sonar:</b> <div style=\"display:inline;position:relative;left:0;top:0;z-index:0\" id=\"sonarstatus\">Unknown</div></td><td>\r\n"
"<input type=\"submit\" value=\"Enable\" onClick=\"sendRequest('/setsonar.bin?e=1')\"><input type=\"submit\" value=\"Disable\" onClick=\"sendRequest('/setsonar.bin?e=0')\"></td></tr><tr><td>\r\n"
"<b>Camera:</b> <div style=\"display:inline;position:relative;left:0;top:0;z-index:0\" id=\"camstatus\">Unknown</div></td><td>\r\n"
"<input type=\"submit\" value=\"Enable\" onClick=\"sendRequest('/setcam.bin?e=1')\"><input type=\"submit\" value=\"Disable\" onClick=\"sendRequest('/setcam.bin?e=0')\"></td></tr><tr><td>\r\n"
"<b>Orders:</b> <div style=\"display:inline;position:relative;left:0;top:0;z-index:0\" id=\"robotorders\">Unknown</div></td><td>\r\n"
"<input type=\"submit\" value=\"Explore\" onClick=\"sendOrders(1)\"><input type=\"submit\" value=\"Wander\" onClick=\"sendOrders(3)\"><input type=\"submit\" value=\"Stop\" onClick=\"sendOrders(4)\"></td></tr><tr><td>\r\n"
"<b>Action:</b> <div style=\"display:inline;position:relative;left:0;top:0;z-index:0\" id=\"robotaction\">Unknown</div></td><td>\r\n"
"</td></tr></table><br>\r\n"
"<div style=\"display:inline;position:absolute;left:8;top:170;z-index:0\" id=\"main\">\r\n"
"<img src=\"map.bmp\" id=\"mapimage\" onClick=\"getCoords(event)\">\r\n"
"<div style=\"display:inline;position:absolute;left:0;top:0;z-index:1\" id=\"robotpos\">\r\n"
"<table border=0 width=10 height=10 bgcolor=\"#ff0000\"><tr><td> </td></tr></table>\r\n"
"</div>\r\n"
"<div style=\"display:inline;position:absolute;left:0;top:0;z-index:1\" id=\"robothead\">\r\n"
"<table border=0 width=6 height=6 bgcolor=\"#ff55ff\"><tr><td> </td></tr></table>\r\n"
"</div>\r\n"
//"<br><img src=\"camimage.jpg\" id=\"camimage\">\r\n"
"</div>\r\n"
"</body>\r\n"
"</html>\r\n"
;
		char headers[128];
		
		sprintf(headers, "HTTP/1.1 200 OK\r\nContent-Type: text/html\r\nContent-Length: %d\r\nConnection: close\r\n\r\n", strlen(content));
		if(m_sock->write(headers, strlen(headers)) != (int)strlen(headers)) {
		}
		if(m_sock->write(content, strlen(content)) != (int)strlen(content)) {
		}
		
		return;
	}
	
	// Send the map-image file map.bmp
	else if(m_requestFile.length() >= 8 && m_requestFile.compare(0, 8, "/map.bmp") == 0) {
		if(m_map->getMapData(m_mapData, &len)) {
			char headers[256];
			
			sprintf(headers, "HTTP/1.1 200 OK\r\nContent-Type: image/bmp\r\nContent-Length: %d\r\nCache-Control: no-store\r\nCache-Control: no-cache\r\nPragma: no-cache\r\nConnection: close\r\n\r\n", len);
			
			if(m_sock->write(headers, strlen(headers)) != (int)strlen(headers)) {
			}
			if(m_sock->write(m_mapData, len) != len) {
			}
			
			//delete[] data;
			
			return;
		}
	}
	
	// temp.bmp is an image of all the sonar-input points (kept by ProbabilityMap, for debugging and inspection)
	else if(m_requestFile.length() >= 9 && m_requestFile.compare(0, 9, "/temp.bmp") == 0) {
		if(m_map->getTempData(m_mapData, &len)) {
			char headers[256];
			
			sprintf(headers, "HTTP/1.1 200 OK\r\nContent-Type: image/bmp\r\nContent-Length: %d\r\nCache-Control: no-store\r\nCache-Control: no-cache\r\nPragma: no-cache\r\nConnection: close\r\n\r\n", len);
			
			if(m_sock->write(headers, strlen(headers)) != (int)strlen(headers)) {
			}
			if(m_sock->write(m_mapData, len) != len) {
			}
			
			//delete[] data;
			
			return;
		}
	}
	
	// Shut the robot down
	else if(m_requestFile.compare(0, m_requestFile.length(), "/shutdown.bin", 13) == 0) {
		const char *content = "<html><head><title>Shutdown</title></head><body bgcolor=\"#ffffff\"><h1>Shutting Down</h1><a href=\"/\">Robot Admin</a><br></body></html>";
		char headers[256];
		
		sprintf(headers, "HTTP/1.1 200 OK\r\nContent-Type: text/html\r\nContent-Length: %d\r\nCache-Control: no-store\r\nCache-Control: no-cache\r\nPragma: no-cache\r\nConnection: close\r\n\r\n", strlen(content));
		
		if(m_sock->write(headers, strlen(headers)) != (int)strlen(headers)) {
		}
		if(m_sock->write(content, strlen(content)) != (int)strlen(content)) {
		}
		
		// Initiate the actual shutdown sequence (global function in RobotAI.cpp)
		doShutdown();
		
		return;
	}
	
	// Send the robot coordinates
	else if(m_requestFile.compare(0, m_requestFile.length(), "/getcoords.bin", 14) == 0) {
		char content[64];
		char headers[256];
		
		sprintf(content, "%d,%d,%d", m_robotHandler->getRobotX(), m_robotHandler->getRobotY(), m_robotHandler->getRobotHeading());
		sprintf(headers, "HTTP/1.1 200 OK\r\nContent-Type: text/html\r\nContent-Length: %d\r\nCache-Control: no-store\r\nCache-Control: no-cache\r\nPragma: no-cache\r\nConnection: close\r\n\r\n", strlen(content));
		
		if(m_sock->write(headers, strlen(headers)) != (int)strlen(headers)) {
		}
		if(m_sock->write(content, strlen(content)) != (int)strlen(content)) {
		}
		
		return;
	}
	
	// Send robot status
	else if(m_requestFile.compare(0, m_requestFile.length(), "/getstatus.bin", 14) == 0) {
		char content[64];
		char headers[256];
		
		sprintf(content, "%s,%s,%s,%s", m_robotHandler->getSonarStatus(), m_robotHandler->getCamStatus(), m_robotHandler->getOrders(), m_robotHandler->getAction());
		sprintf(headers, "HTTP/1.1 200 OK\r\nContent-Type: text/html\r\nContent-Length: %d\r\nCache-Control: no-store\r\nCache-Control: no-cache\r\nPragma: no-cache\r\nConnection: close\r\n\r\n", strlen(content));
		
		if(m_sock->write(headers, strlen(headers)) != (int)strlen(headers)) {
		}
		if(m_sock->write(content, strlen(content)) != (int)strlen(content)) {
		}
		
		return;
	}
	
	// Send latest camera image (as JPEG)
	else if(m_requestFile.length() >= 13 && m_requestFile.compare(0, 13, "/camimage.jpg") == 0) {
		CamImage *img = m_imgProvider->getImage();
		if(img) {
			char headers[256];
			
			sprintf(headers, "HTTP/1.1 200 OK\r\nContent-Type: image/jpeg\r\nContent-Length: %d\r\nCache-Control: no-store\r\nCache-Control: no-cache\r\nPragma: no-cache\r\nConnection: close\r\n\r\n", img->dataLen);
			
			if(m_sock->write(headers, strlen(headers)) != (int)strlen(headers)) {
			}
			if(m_sock->write(img->imageData, img->dataLen) != img->dataLen) {
			}
			
			img->release();
			
			return;
		}
	}
	
	// Send latest camera image (as BMP)
	else if(m_requestFile.length() >= 13 && m_requestFile.compare(0, 13, "/camimage.bmp") == 0) {
		CamImage *img = m_imgProvider->getImage();
		if(img) {
			bool go = img->decodeImage();
			if(!go) {
				ArUtil::sleep(100);
				go = img->decodeImage();
			}
			
			if(go) {
				go = img->makeBmp();
				if(!go) {
					ArUtil::sleep(100);
					go = img->makeBmp();
				}
				
				if(go) {
					char headers[256];
					
					sprintf(headers, "HTTP/1.1 200 OK\r\nContent-Type: image/bmp\r\nContent-Length: %d\r\nCache-Control: no-store\r\nCache-Control: no-cache\r\nPragma: no-cache\r\nConnection: close\r\n\r\n", img->bmpLen);
					
					if(m_sock->write(headers, strlen(headers)) != (int)strlen(headers)) {
					}
					if(m_sock->write(img->bmp, img->bmpLen) != img->bmpLen) {
					}
				}
			}
			
			img->release();
			
			return;
		}
	}
	
	// Auto-updating java-script page showing the camera image as BMP
	// (The BMP in this version has a grid drawn on it when built in the CamImage class,
	// so it is good to use for distance-estimates)
	else if(m_requestFile.compare(0, m_requestFile.length(), "/camimage.html", 14) == 0) {
		const char *content =
"<html>\r\n"
"<head>\r\n"
"<title>Cam Image</title>\r\n"
"<script language=\"JavaScript\">\r\n"
"function reloadImg() {\r\n"
"	document.getElementById(\"camimage\").src = \"camimage.bmp?\" + (new Date()).getTime();\r\n"
"	setTimeout('reloadImg()', 5000);\r\n"
"}\r\n"
"</script>\r\n"
"</head>\r\n"
"<body onLoad=\"reloadImg()\">\r\n"
"<img src=\"camimage.bmp\" width=960 height=720 id=\"camimage\">\r\n"
"</body>\r\n"
"</html>\r\n"
;
		char headers[128];
		
		sprintf(headers, "HTTP/1.1 200 OK\r\nContent-Type: text/html\r\nContent-Length: %d\r\nConnection: close\r\n\r\n", strlen(content));
		if(m_sock->write(headers, strlen(headers)) != (int)strlen(headers)) {
		}
		if(m_sock->write(content, strlen(content)) != (int)strlen(content)) {
		}
		
		return;
	}
	
	// Handle order to move the robot to a specified position
	else if(m_requestFile.length() > movetoLen && m_requestFile.compare(0, movetoLen, movetoFile) == 0) {
		const char *content = "<html><head><title>COORDS!</title></head><body><h1>COORDS!</h1></body></html>";
		char headers[256];
		
		sprintf(headers, "HTTP/1.1 200 OK\r\nContent-Type: text/html\r\nContent-Length: %d\r\nCache-Control: no-store\r\nCache-Control: no-cache\r\nPragma: no-cache\r\nConnection: close\r\n\r\n", strlen(content));
		
		if(m_sock->write(headers, strlen(headers)) != (int)strlen(headers)) {
		}
		if(m_sock->write(content, strlen(content)) != (int)strlen(content)) {
		}
		
		// Parse the move-to coords
		parseMoveto(m_requestFile.c_str(), static_cast<int>(m_requestFile.length()));
		
		return;
	}
	
	// Handle orders
	else if(m_requestFile.length() > orderLen && m_requestFile.compare(0, orderLen, orderFile) == 0) {
		const char *content = "<html><head><title>ORDERS!</title></head><body><h1>ORDERS!</h1></body></html>";
		char headers[256];
		
		sprintf(headers, "HTTP/1.1 200 OK\r\nContent-Type: text/html\r\nContent-Length: %d\r\nCache-Control: no-store\r\nCache-Control: no-cache\r\nPragma: no-cache\r\nConnection: close\r\n\r\n", strlen(content));
		
		if(m_sock->write(headers, strlen(headers)) != (int)strlen(headers)) {
		}
		if(m_sock->write(content, strlen(content)) != (int)strlen(content)) {
		}
		
		// Parse the orders
		parseOrders(m_requestFile.c_str(), static_cast<int>(m_requestFile.length()));
		
		return;
	}
	
	// Handle order to enable/disable camera
	else if(m_requestFile.length() > camLen && m_requestFile.compare(0, camLen, camFile) == 0) {
		const char *content = "<html><head><title>CAM!</title></head><body><h1>CAM!</h1></body></html>";
		char headers[256];
		
		sprintf(headers, "HTTP/1.1 200 OK\r\nContent-Type: text/html\r\nContent-Length: %d\r\nCache-Control: no-store\r\nCache-Control: no-cache\r\nPragma: no-cache\r\nConnection: close\r\n\r\n", strlen(content));
		
		if(m_sock->write(headers, strlen(headers)) != (int)strlen(headers)) {
		}
		if(m_sock->write(content, strlen(content)) != (int)strlen(content)) {
		}
		
		// Parse camera settings
		parseCam(m_requestFile.c_str(), static_cast<int>(m_requestFile.length()));
		
		return;
	}
	
	// Handle order to enable/disable sonars
	else if(m_requestFile.length() > sonarLen && m_requestFile.compare(0, sonarLen, sonarFile) == 0) {
		const char *content = "<html><head><title>SONAR!</title></head><body><h1>SONAR!</h1></body></html>";
		char headers[256];
		
		sprintf(headers, "HTTP/1.1 200 OK\r\nContent-Type: text/html\r\nContent-Length: %d\r\nCache-Control: no-store\r\nCache-Control: no-cache\r\nPragma: no-cache\r\nConnection: close\r\n\r\n", strlen(content));
		
		if(m_sock->write(headers, strlen(headers)) != (int)strlen(headers)) {
		}
		if(m_sock->write(content, strlen(content)) != (int)strlen(content)) {
		}
		
		// Parse sonar settings
		parseSonar(m_requestFile.c_str(), static_cast<int>(m_requestFile.length()));
		
		return;
	}
	
	// Default 404 message
	{
		const char *content = "<html><head><title>404 Not Found</title></head><body><h1>404 Not Found</h1></body></html>";
		char headers[128];
		
		sprintf(headers, "HTTP/1.1 404 Not Found\r\nContent-Type: text/html\r\nContent-Length: %u\r\nConnection: close\r\n\r\n", strlen(content));
		
		if(m_sock->write(headers, strlen(headers)) != (int)strlen(headers)) {
		}
		if(m_sock->write(content, strlen(content)) != (int)strlen(content)) {
		}
	}
}// HTTPConnection::sendFile


// Parse moveto request
// moveto.bin?x=XPOSITION;y=YPOSITION
// (There's probably a standard-library thingy for parsing this easier..)
void HTTPConnection::parseMoveto(const char *str, int len) {
	int x = -1;
	int y = -1;
	
	*(char*)(str+len) = 0;
	//printf("moveto: \"%s\"\n", str);
	
	int i = 1;
	for(;i<len;i++) {
		if(str[i-1] == 'x' && str[i] == '=') {
			for(int j=i+1;j<len;j++) {
				if(str[j] >= '0' && str[j] <= '9') {
				}
				else if(str[j] == ';' && j > i+1 && j < i+10) {
					char anum[10];
					memcpy(anum, str+i+1, j-i-1);
					anum[j-i-1] = 0;
					x = atoi(anum);
					break;
				}
				else break;
			}
			break;
		}
	}
	
	for(;i<len;i++) {
		if(str[i-1] == 'y' && str[i] == '=') {
			for(int j=i+1;j<len;j++) {
				if(str[j] >= '0' && str[j] <= '9') {
					if(j == len-1) {
						char anum[10];
						anum[j-i] = 0;
						memcpy(anum, str+i+1, j-i);
						y = atoi(anum);
						break;
					}
				}
				else if(str[j] == ';' && j > i+1 && j < i+10) {
					char anum[10];
					anum[j-i-1] = 0;
					memcpy(anum, str+i+1, j-i-1);
					y = atoi(anum);
					break;
				}
				else break;
			}
			break;
		}
	}
	
	if(x >= 0 && y >= 0) {
		//printf("x = %d\ny = %d\n", x, y);
		m_robotHandler->gotoMapPos(x, y);
	}
}// HTTPConnection::parseMoveTo


// Parse orders request
// o=X, where X is the order number
// Orders are just stored as a number
void HTTPConnection::parseOrders(const char *str, int len) {
	int o = -1;
	
	int i = 1;
	for(;i<len;i++) {
		if(str[i-1] == 'o' && str[i] == '=') {
			for(int j=i+1;j<len;j++) {
				if(str[j] >= '0' && str[j] <= '9') {
					if(j == len-1) {
						char anum[10];
						anum[j-i] = 0;
						memcpy(anum, str+i+1, j-i);
						o = atoi(anum);
						break;
					}
				}
				else break;
			}
			break;
		}
	}
	
	if(o >= 0) {
		m_robotHandler->inputOrders(o);
	}
}// HTTPConnection::parseOrders


// Parse cam request
// e=0/1
void HTTPConnection::parseCam(const char *str, int len) {
	int e = -1;
	
	int i = 1;
	for(;i<len;i++) {
		if(str[i-1] == 'e' && str[i] == '=') {
			for(int j=i+1;j<len;j++) {
				if(str[j] >= '0' && str[j] <= '1') {
					if(j == len-1) {
						char anum[10];
						anum[j-i] = 0;
						memcpy(anum, str+i+1, j-i);
						e = atoi(anum);
						break;
					}
				}
				else break;
			}
			break;
		}
	}
	
	if(e == 0) {
		m_robotHandler->enableCam(false);
	}
	else if(e == 1) {
		m_robotHandler->enableCam(true);
	}
}// HTTPConnection::parseCam



// Parse sonar request
// e=0/1
void HTTPConnection::parseSonar(const char *str, int len) {
	int e = -1;
	
	int i = 1;
	for(;i<len;i++) {
		if(str[i-1] == 'e' && str[i] == '=') {
			for(int j=i+1;j<len;j++) {
				if(str[j] >= '0' && str[j] <= '1') {
					if(j == len-1) {
						char anum[10];
						anum[j-i] = 0;
						memcpy(anum, str+i+1, j-i);
						e = atoi(anum);
						break;
					}
				}
				else break;
			}
			break;
		}
	}
	
	if(e == 0) {
		m_robotHandler->enableSonar(false);
	}
	else if(e == 1) {
		m_robotHandler->enableSonar(true);
	}
}// HTTPConnection::parseSonar
