// ImageProvider.h

// The ImageProvider is an asynchronous task that connects to the camera
// and receives images from it continously
// Other threads can request an image from the ImageProvider, which returns
// the latest image as a CamImage (which is reference counted, so auto-deleted
// when the last reference is removed)

#include "ImageProvider.h"
#include "CamImage.h"
#include "strstuff.h"// String functions, should probably be replaced with stdlib



// Construct
ImageProvider::ImageProvider() :
	m_curImage(0),
	
	m_nextImageId(1),
	
	m_data(0),
	m_dataLen(0),
	
	m_state(0),
	m_contentLength(0)
{
}// ImageProvider::ImageProvider


// Destruct
ImageProvider::~ImageProvider() {
	if(m_curImage)
		m_curImage->release();
}// ImageProvider::~ImageProvider


// Main thread
void* ImageProvider::runThread(void*) {
	// Connect
	//if(m_sock.connect("130.235.21.161", 80, ArSocket::TCP)) {// WIFI camera
	if(m_sock.connect("192.168.0.2", 80, ArSocket::TCP)) {// Ethernet camera
		// Request
		//const char *get = "GET /mjpg/video.mjpg HTTP/1.0\r\nHost: 130.235.21.161\r\n\r\n";
		const char *get = "GET /mjpg/video.mjpg HTTP/1.0\r\nHost: 192.168.0.2\r\n\r\n";
		
		if(m_sock.write(get, strlen(get)) == (int)strlen(get)) {
			const int maxRead = 256*1024;
			char *data = new char[maxRead];
			
			// Loop
			while(getRunningWithLock()) {
				int len = m_sock.read(data, maxRead);
				if(len > 0) {
					if(handleData(data, len)) {
						break;
					}
				}
				else {
					break;
				}
			}
			
			delete[] data;
		}
		
		m_sock.close();
	}
	
	return NULL;
}// ImageProvider::runThread


// Close the socket
void ImageProvider::myStop() {
	m_sock.close();
}// ImageProvider::myStop


// Handle data
// Parse incoming data to images
bool ImageProvider::handleData(char *data, int len) {
	bool done = false;
	char *oldData = m_data;
	
	m_data = new char[m_dataLen+len];
	memcpy(m_data, oldData, m_dataLen);
	memcpy(m_data+m_dataLen, data, len);
	m_dataLen += len;
	delete[] oldData;
	
	// Images will be continously returned
	// Check the AXIS camera manual for description of the headers etc. involved
	// which are parsed here
	bool rego = true;
	while(rego && m_dataLen > 0) {
		rego = false;
		
		switch(m_state) {
			// Look for header
			case 0:
				for(int i=3;i<m_dataLen;i++) {
					if(m_data[i-3] == '\r' && m_data[i-2] == '\n' && m_data[i-1] == '\r' && m_data[i] == '\n') {
						memmove(m_data, m_data+i+1, m_dataLen-(i+1));
						m_dataLen -= (i+1);
						
						m_state = 1;
						rego = true;
					}
				}
			break;
			
			// Look for content
			case 1:
				for(int i=3;i<m_dataLen;i++) {
					if(m_data[i-3] == '\r' && m_data[i-2] == '\n' && m_data[i-1] == '\r' && m_data[i] == '\n') {
						m_contentLength = parseContentLength(m_data, i+1);
						if(m_contentLength > 0) {
							memmove(m_data, m_data+i+1, m_dataLen-(i+1));
							m_dataLen -= (i+1);
							
							m_state = 2;
							rego = true;
						}
						else {
							done = true;
						}
					}
				}
			break;
			
			// Actual image data
			case 2:
				if(m_dataLen >= m_contentLength) {
					oldData = m_data;
					if(m_dataLen > m_contentLength) {
						m_data = new char[m_dataLen-m_contentLength];
						memcpy(m_data, oldData+m_contentLength, m_dataLen-m_contentLength);
						
						rego = true;
					}
					else {
						m_data = 0;
					}
					m_dataLen -= m_contentLength;
					
					m_state = 0;
					
					// Construct the image
					CamImage *newImage = new CamImage(m_nextImageId++, oldData, m_contentLength);
					CamImage *oldImage;
					
					// Replace the old image (thread-safely)
					m_mutex.lock();
					oldImage = m_curImage;
					m_curImage = newImage;
					m_mutex.unlock();
					
					// Release the old image (will be deleted when reference-count reaches 0)
					// (which might be now, or later if another thread has a reference)
					if(oldImage)
						oldImage->release();
				}
			break;
		}
	}
	
	return done;
}// ImageProvider::handleData


// Parse content length header
int ImageProvider::parseContentLength(const char *data, int len) {
	int contentLength = 0;
	int lineLen;
	int next = 0;
	
	while(contentLength == 0 && (lineLen = getLine(data, len, next)) > 2) {
		contentLength = findContentLength(data+next, lineLen-2);
		
		next += lineLen;
	}
	
	return contentLength;
}// ImageProvider::parseContentLength


// Find content length in line of HTTP header
// This could use the standard library instead of custom functions..
int ImageProvider::findContentLength(const char *line, int len) {
	int partLen;
	int next = 0;
	
	partLen = findChar(line, len, ':', next);
	if(partLen > 0) {
		if(strEqC(line+next, partLen, "content-length")) {
			next += partLen + 1;
			
			partLen = findNotChar(line, len, ' ', next);
			next += partLen;
			
			partLen = findChar(line, len, ';', next);
			if(partLen > 0) {
				char str[64];
				
				memcpy(str, line+next, partLen);
				str[partLen] = 0;
				
				int clen = atoi(str);
				
				if(clen > 0) {
					return clen;
				}
			}
		}
	}
	
	return 0;
}// ImageProvider::parseContentLength


// Get latest image
CamImage* ImageProvider::getImage() {
	CamImage *img = 0;
	
	m_mutex.lock();
	if(m_curImage)
		img = m_curImage->acquire();
	m_mutex.unlock();
	
	return img;
}// ImageProvider::getImage
