// LocalizationMap.cpp

// The LocalizationMap records sonar input points, and searches for
// lines in the data
// When lines are found their angle is calculated, and returned
// It is used to find when the robot's percieved movement angle is
// different from its true angle, because of odometry drift
// Walls are estimated to be at 90-degree angles, which is why
// the angle of the lines can be used for this purpose

#include "LocalizationMap.h"
#include "CamImage.h"

#include "Aria.h"
#include <memory>


// Construct
LocalizationMap::LocalizationMap() :
	// Fixed-size map, perhaps this should be fixed
	m_pixels(0),
	m_width(768),
	m_height(768),
	
	m_points(0),
	m_nPoints(0),
	m_nMaxPoints(400),// Points to wait for before searching for lines
	
	m_lvals(0),
	m_lwidth(100),
	m_lheight(100),
	m_lrscale(0.5),
	m_ctable(0),
	m_stable(0),
	
	m_lastt(0.0)
{
	if(m_width > 0 && m_height > 0) {
		m_pixels = new unsigned char[m_height*m_width];
		for(int i=0;i<m_height*m_width;i++)
			m_pixels[i] = 255;
	}
	
	m_points = new vec2d[m_nMaxPoints];
	
	if(m_lwidth > 0 && m_lheight > 0) {
		m_lvals = new unsigned int[m_lheight*m_lwidth];
		for(int i=0;i<m_lheight*m_lwidth;i++)
			m_lvals[i] = 0;
		
		m_ctable = new double[m_lheight];
		m_stable = new double[m_lheight];
		for(int i=0;i<m_lheight;i++) {
			double a = 3.14159265358979 * static_cast<double>(i) / static_cast<double>(m_lheight);
			m_ctable[i] = cos(a);
			m_stable[i] = sin(a);
		}
	}
}// LocalizationMap::LocalizationMap


// Destruct
LocalizationMap::~LocalizationMap() {
	for(std::vector<Line*>::iterator i=m_segments.begin();i!=m_segments.end();++i)
		delete (*i);
	
	delete[] m_pixels;
	
	delete[] m_points;
	delete[] m_lvals;
	delete[] m_ctable;
	delete[] m_stable;
}// LocalizationMap::~LocalizationMap


// Process input
bool LocalizationMap::input(const vec2d &pos, const Ray *rays, int nRays) {
	bool drift = false;
	
	// Record all rays (rays are from the robot to the sonar-detection)
	for(int i=0;i<nRays;i++) {
		const Ray &r(rays[i]);
		
		if(r.valid) {
			if(r.endHit) {
				double scale = 1.0 / 50.0;
				
				//int x = static_cast<int>(r.p1.x * scale) + m_width/2;
				//int y = static_cast<int>(r.p1.y * scale) + m_height/2;
				
				//m_pixels[y*m_width+x] = 128;
				
				if(m_nPoints < m_nMaxPoints) {
					m_points[m_nPoints] = vec2d(r.p1.x * scale, r.p1.y * scale);
					m_nPoints++;
				}
			}
		}
	}
	
	// Find lines when enough points are recorded
	if(m_nPoints == m_nMaxPoints) {
		findLines();
		drift = true;
	}
	
	return drift;
}// LocalizationMap::input



// Get theta

double LocalizationMap::getTheta() {
	return m_lastt;
}// LocalizationMap::getTheta


// Find lines
// Uses the Hough transform to find lines as y = mx + b
void LocalizationMap::findLines() {
	//for(size_t i=0;i<m_segments.size();i++)
	//	delete m_segments[i];
	//m_segments.clear();
	
	// Uses very rough estimate, and simply chooses the line
	// with the most number of points in it
	// Should really search for more lines and compare them,
	// if better estimates are to be obtained
	m_lastt = 0.0;
	
	for(int i=0;i<m_lheight;i++) {
		for(int j=0;j<m_lwidth;j++) {
			m_lvals[i*m_lwidth+j] = 0;
		}
	}
	
	double minx = 99999.0, miny = 99999.0, maxx = -99999.0, maxy = -99999.0;
	for(int i=0;i<m_nPoints;i++) {
		const vec2d &p(m_points[i]);
		
		if(p.x < minx) minx = p.x;
		if(p.y < miny) miny = p.y;
		if(p.x > maxx) maxx = p.x;
		if(p.y > maxy) maxy = p.y;
	}
	for(int i=0;i<m_nPoints;i++) {
		m_points[i].x -= minx;
		m_points[i].y -= miny;
	}
	
	for(int i=0;i<m_nPoints;i++) {
		const vec2d &p(m_points[i]);
		
		for(int j=0;j<m_lheight;j++) {
			double r = p.x * m_ctable[j] + p.y * m_stable[j];
			int ir = static_cast<int>(r  * m_lrscale) + (m_lwidth/2);
			
			if(ir >= 0 && ir < m_lwidth) {
				m_lvals[j*m_lwidth+ir] += 3;
			}
		}
	}
	
	unsigned int maxval = 0;
	double t, r, m, b;
	for(int i=0;i<m_lheight;i++) {
		for(int j=0;j<m_lwidth;j++) {
			if(m_lvals[i*m_lwidth+j] > maxval) {
				maxval = m_lvals[i*m_lwidth+j];
				t = 3.14159265358979 * static_cast<double>(i) / static_cast<double>(m_lheight);
				r = static_cast<double>(j - (m_lwidth/2)) / m_lrscale;
				m = -cos(t) / sin(t);
				b = r / sin(t);
			}
		}
	}
	
	if(maxval > 0) {
		// For debug, view the lines //
		/*Line *l = new Line;
		
		l->line.p0.x = -100.0 * -sin(t);
		l->line.p0.y = -100.0 * cos(t);
		l->line.p1.x = 100.0 * -sin(t);
		l->line.p1.y = 100.0 * cos(t);
		
		l->line.p0.x += r * cos(t) + minx;
		l->line.p0.y += r * sin(t) + miny;
		l->line.p1.x += r * cos(t) + minx;
		l->line.p1.y += r * sin(t) + miny;
		
		l->line.p0 *= 50.0;
		l->line.p1 *= 50.0;
		
		m_segments.push_back(l);*/
		
		m_lastt = t;
	}
	
	m_nPoints = 0;
}// LocalizationMap::findLines


// Draw grayscale map for viewing (debug)
void LocalizationMap::drawGrayscale(unsigned char *data, int w, int h, double scale) {
	for(int i=0;i<h&&i<m_height;i++) {
		for(int j=0;j<w&&j<m_width;j++) {
			*(data + i*w + j) = *(m_pixels + i*m_width + j);
		}
	}
	
	for(std::vector<Line*>::iterator i=m_segments.begin();i!=m_segments.end();++i) {
		const Line *ls = *i;
		
		drawLine(data, w, h, ls->line*scale);
	}
}// LocalizationMap::drawGrayscale


// Draw line
// Trivial line drawing of floating point edges (debug)
void LocalizationMap::drawLine(unsigned char *data, int w, int h, const line2d &seg) {
	vec2d l(seg.p1-seg.p0);
	
	double dw = static_cast<double>(w/2);
	double dh = static_cast<double>(h/2);
	
	if(!(seg.p0.x > -dw && seg.p0.x < dw && seg.p1.x > -dw && seg.p1.x < dw &&
		seg.p0.y > -dh && seg.p0.y < dh && seg.p1.y > -dh && seg.p1.y < dh)) {
		return;
	}
	
	if(l.x*l.x > l.y*l.y) {
		if(l.x > 0.0) {
			for(double x=0.0;x<l.x;x+=1.0) {
				double y = seg.p0.y + l.y * x / l.x;
				
				int ix = static_cast<int>(x + seg.p0.x) + w/2;
				int iy = static_cast<int>(y) + h/2;
				
				*(data + iy*w + ix) = 0;
			}
		}
		else {
			for(double x=0.0;x>l.x;x-=1.0) {
				double y = seg.p0.y + l.y * x / l.x;
				
				int ix = static_cast<int>(x + seg.p0.x) + w/2;
				int iy = static_cast<int>(y) + h/2;
				
				*(data + iy*w + ix) = 0;
			}
		}
	}
	else {
		if(l.y > 0.0) {
			for(double y=0.0;y<l.y;y+=1.0) {
				double x = seg.p0.x + l.x * y / l.y;
				
				int ix = static_cast<int>(x) + w/2;
				int iy = static_cast<int>(y + seg.p0.y) + h/2;
				
				*(data + iy*w + ix) = 0;
			}
		}
		else {
			for(double y=0.0;y>l.y;y-=1.0) {
				double x = seg.p0.x + l.x * y / l.y;
				
				int ix = static_cast<int>(x) + w/2;
				int iy = static_cast<int>(y + seg.p0.y) + h/2;
				
				*(data + iy*w + ix) = 0;
			}
		}
	}
}// LocalizationMap::drawLine
