// ProbabilityMap.cpp

// This is the main probability map of the robot's environment
// It keeps track of the map as a grid where each cell has a probability
// of being obstructed.
// It also handles path-finding etc.

// This class is too big and should be refactored..
// It handles things that probably shouldn't be its responsibility

// Cells are kept as 32-bit values
// Only the low 8 bits are used for occupancy probability
// Bit 9 is whether the cell is unknown to the robot or has been explored
// at least once
// Higher bits are currently unused

#include "ProbabilityMap.h"
#include "Aria.h"


// Construct
ProbabilityMap::ProbabilityMap(int w, int h) :
	m_cells(0),// Occupancy information
	m_tempCells(0),// Temporary cell storage
	m_costs(0),// Cell costs for pathfinding
	m_prev(0),// Used to remember paths when pathfinding
	m_width(w),
	m_height(h),
	
	m_hasRays(false),
	
	m_nearSize(10, 10),
	m_veryNearSize(5, 5),
	m_rayWidth(5.0),
	m_rayRetract(5.0),
	m_hitSize(5, 5),
	m_largeHitSize(10, 10),
	m_maxPassableCost(200),
	
	m_path(0),
	m_pathLen(0),
	
	m_hasTotalRect(false)
{
	// The map cells
	if(m_width > 0 && m_height > 0) {
		m_cells = new unsigned int[m_height*m_width];
		for(int i=0;i<m_height*m_width;i++)
			m_cells[i] = 100 | 0x100;
		
		m_tempCells = new unsigned int[m_height*m_width];
		for(int i=0;i<m_height*m_width;i++)
			m_tempCells[i] = 0xff;
		
		m_costs = new unsigned int[m_height*m_width];
		for(int i=0;i<m_height*m_width;i++)
			m_costs[i] = 0;
		
		m_prev = new unsigned int[m_height*m_width];
		for(int i=0;i<m_height*m_width;i++)
			m_prev[i] = 0;
	}
	
	// Pathfinding path (store it for viewing, it will be put in the image
	// when an image of the map is requested in ProbabilityMapImage.cpp)
	const int maxPathLen = 1000;
	m_path = new vec2i[maxPathLen];
}// ProbabilityMap::ProbabilityMap


// Destruct
ProbabilityMap::~ProbabilityMap() {
	delete[] m_cells;
	delete[] m_tempCells;
	delete[] m_costs;
	delete[] m_prev;
}// ProbabilityMap::~ProbabilityMap


// Input rays from sonars
void ProbabilityMap::input(const vec2d &pos, const Ray *rays, int nRays) {
	int minx = 99999, miny = 99999, maxx = -99999, maxy = -99999;
	
	// Use last iteration rays together with new rays
	if(m_hasRays) {
		for(int i=0;i<nRays;i++) {
			const Ray &ray(rays[i]);
			
			if(ray.valid && m_rays[i].valid) {
				// Decrease the probability of obstacles in the proximity of the
				// line traveled by the sonar
				fillLine2(ray, m_rays[i], m_rayWidth, 2, 50);
				
				// Record obstacles if the sonar hit something
				if(ray.endHit && m_rays[i].endHit) {
					fillRect2(ray.p1, m_rays[i].p1, vec2d(7.0, 3.5), 255);
					
					m_tempCells[((int)(ray.p1.y))*m_width+(int)ray.p1.x] = 0;
				}
				
				// Update rect for dirty rects
				int x, y;
				
				x = static_cast<int>(ray.p0.x);
				y = static_cast<int>(ray.p0.y);
				if(x < minx) minx = x;
				if(x > maxx) maxx = x;
				if(y < miny) miny = y;
				if(y > maxy) maxy = y;
				x = static_cast<int>(ray.p1.x);
				y = static_cast<int>(ray.p1.y);
				if(x < minx) minx = x;
				if(x > maxx) maxx = x;
				if(y < miny) miny = y;
				if(y > maxy) maxy = y;
			}
		}
	}
	
	// Store rays for next iteration
	for(int i=0;i<nRays;i++)
		m_rays[i] = rays[i];
	m_hasRays = true;
	
	// Make sure grid very close to robot is open
	fillRect(vec2dtoi(pos), m_nearSize, 0, 101);
	
	// Add dirty rect of the part of the map that is changed
	// (for updates to network viewers)
	int x, y;
	
	x = static_cast<int>(pos.x);
	y = static_cast<int>(pos.y);
	if(x < minx) minx = x;
	if(x > maxx) maxx = x;
	if(y < miny) miny = y;
	if(y > maxy) maxy = y;
	
	Rect rect;
	
	rect.left = minx;
	rect.top = miny;
	rect.width = maxx - minx;
	rect.height = maxy - miny;
	
	addDirtyRect(rect);
}// ProbabilityMap::input


// Get map dimensions
int ProbabilityMap::width() const {
	return m_width;
}// ProbabilityMap::width
int ProbabilityMap::height() const {
	return m_height;
}// ProbabilityMap::height
vec2d ProbabilityMap::mapCenter() const {
	return vec2d(static_cast<double>(m_width/2), static_cast<double>(m_height/2));
	//return vec2d((double)(m_width/8), (double)(m_height/2));
}// ProbabilityMap::mapCenter


// Add dirty rect
void ProbabilityMap::addDirtyRect(const Rect &nrect) {
	bool added = false;
	
	// Total used map rect
	if(m_hasTotalRect) {
		const Rect &rect(m_totalRect);
		Rect trect;
		
		trect.left = rect.left < nrect.left ? rect.left : nrect.left;
		trect.top = rect.top < nrect.top ? rect.top : nrect.top;
		trect.width = (rect.width+rect.left-trect.left) > (nrect.width+nrect.left-trect.left) ? (rect.width+rect.left-trect.left) : (nrect.width+nrect.left-trect.left);
		trect.height = (rect.height+rect.top-trect.top) > (nrect.height+nrect.top-trect.top) ? (rect.height+rect.top-trect.top) : (nrect.height+nrect.top-trect.top);
		
		m_totalRect = trect;
	}
	else {
		m_totalRect = nrect;
		m_hasTotalRect = true;
	}
}// ProbabilityMap::addDirtyRect


// Get map rect
Rect ProbabilityMap::getTotalRect() const {
	return m_totalRect;
}// ProbabilityMap::getTotalRect
