// ProbabilityMap.cpp

// ProbabilityMap member functions for filling cells in the map
// Each cell is an unsigned int 32-bit value. The lower 8 bits
// is the probability that the cell is occupied, in [0,255]
// Bit 9 means whether the cell is unknown or not
// Higher bits are unused, but preserved for future purposes

// So when a sonar ray makes an estimation of a new probability for
// a cell, its new value is set to:
// newValue = (oldValue & 0xfffffe00) | newProbability
// This clears the unknown bit 9, so that the cell is considered known
// to have this new probability, and the actual probability is set

// When probability is meant to decrease, then the lower 8 bits are
// read, decreased by 1, clamped to 0, and written back

#include "ProbabilityMap.h"



// Fill a rect in the map with a set probability-value
void ProbabilityMap::fillRect(
	const vec2i &center,
	const vec2i &size,
	unsigned int val
) {
	for(int i=center.y-size.y;i<center.y+size.y;i++) {
		for(int j=center.x-size.x;j<center.x+size.x;j++) {
			m_cells[i*m_height+j] = (m_cells[i*m_height+j] & 0xfffffe00) | val;
		}
	}
}// ProbabilityMap::fillRect


// Fill a rect in the map with a set probability-value, if the previous
// value is less than clipVal
void ProbabilityMap::fillRect(
	const vec2i &center,
	const vec2i &size,
	unsigned int val,
	unsigned int clipVal
) {
	for(int i=center.y-size.y;i<center.y+size.y;i++) {
		for(int j=center.x-size.x;j<center.x+size.x;j++) {
			if((m_cells[i*m_height+j] & 0xff) <= clipVal)
				m_cells[i*m_height+j] = (m_cells[i*m_height+j] & 0xfffffe00) | val;
		}
	}
}// ProbabilityMap::fillRect


// Fill rect
// Two consecutive hits by the same sonar, fill a rectangle stretched towards both
// This is to stretch the blocks in the direction of walls, since two consecutive rays
// from the same sonar are likely to have hit a straight line wall
void ProbabilityMap::fillRect2(
	const vec2d &p0,
	const vec2d &p1,
	const vec2d &size,
	unsigned int val
) {
	vec2d points[4];
	vec2d spoints[4];
	vec2d dir = (p1 - p0);
	
	if(dir.length2() < 0.1) {// To close, just fill a square
		fillRect(vec2dtoi(p0), vec2i(3, 3), val);
		return;
	}
	
	vec2d c = cross(dir).normalize();
	dir.normalize();
	
	// Thick line points
	points[0] = p1 + c * size.y - dir * size.x;
	points[1] = p1 - c * size.y - dir * size.x;
	points[2] = p1 - c * size.y + dir * size.x;
	points[3] = p1 + c * size.y + dir * size.x;
	
	// Sort for poly-filling (counter-clockwise, first point smallest y-val)
	int ifirst = 0;
	for(int i=1;i<4;i++) {
		if(points[i].y < points[ifirst].y)
			ifirst = i;
	}
	for(int i=0;i<4;i++) {
		spoints[i] = points[(ifirst+i)%4];
	}
	
	// Fill
	fillPoly(spoints, 4, val);
}// ProbabilityMap::fillRect2


// Fill line, use two consecutive rays from the same sonar
// to fill a shape between them (+ thickness)
// This method can be used to change the shape of sonar-rays
void ProbabilityMap::fillLine2(
	const Ray &ray0,
	const Ray &ray1,
	double width,
	unsigned int weight,
	unsigned int minVal
) {
	vec2d points[4];
	
	vec2d p0 = (ray0.p0 + ray1.p0) * 0.5;
	vec2d p1 = (ray0.p1 + ray1.p1) * 0.5;
	vec2d c = cross(p1-p0).normalize();
	
	points[0] = p0 + c * width;
	points[1] = p0 - c * width;
	
	double d = dot(p1, c);
	double d0 = dot(ray0.p1, c) - d;
	double d1 = dot(ray1.p1, c) - d;
	
	if(d0 < d1) {
		vec2d c2 = (ray1.p1 - ray0.p1);
		if(c2.length2() > 0.5)
			c = c2.normalize();
		points[2] = ray0.p1 - c * width;
		points[3] = ray1.p1 + c * width;
	}
	else {
		vec2d c2 = (ray0.p1 - ray1.p1);
		if(c2.length2() > 0.5)
			c = c2.normalize();
		points[2] = ray1.p1 - c * width;
		points[3] = ray0.p1 + c * width;
	}
	
	vec2d spoints[4];
	
	// Sort
	int ifirst = 0;
	for(int i=1;i<4;i++) {
		if(points[i].y < points[ifirst].y)
			ifirst = i;
	}
	for(int i=0;i<4;i++) {
		spoints[i] = points[(ifirst+i)%4];
	}
	
	// Fill
	fillPoly(spoints, 4, weight, minVal);
}// ProbabilityMap::fillLine2


// Fill polygon area, by decreasing occupancy probability for the grid cells in the poly
void ProbabilityMap::fillPoly(
	const vec2d *points,
	int nPoints,
	unsigned int weight,
	unsigned int minVal
) {
	int iPoint = 0;
	int iLeft;
	int iRight;
	double dx[2], dax[2];
	double dy;
	int iy;
	
	minVal = 0;
	
	iy = static_cast<int>(points[iPoint].y + 0.5);
	dy = static_cast<double>(iy) + 0.5;
	
	iLeft = 1;
	iRight = nPoints-1;
	
	dax[0] = (points[iLeft].x - points[iPoint].x) / (points[iLeft].y - points[iPoint].y);
	dx[0] = points[iPoint].x + (dy - points[iPoint].y) * dax[0];
	
	dax[1] = (points[iRight].x - points[iPoint].x) / (points[iRight].y - points[iPoint].y);
	dx[1] = points[iPoint].x + (dy - points[iPoint].y) * dax[1];
	
	for(;;) {
		dy = static_cast<double>(iy) + 0.5;
		
		while(points[iLeft].y < dy) {
			iLeft++;
			if(iLeft > iRight)
				break;
			
			iPoint = iLeft-1;
			
			dax[0] = (points[iLeft].x - points[iPoint].x) / (points[iLeft].y - points[iPoint].y);
			dx[0] = points[iPoint].x + (dy - points[iPoint].y) * dax[0];
		}
		
		while(points[iRight].y < dy) {
			iRight--;
			if(iLeft > iRight)
				break;
			
			iPoint = iRight+1;
			
			dax[1] = (points[iRight].x - points[iPoint].x) / (points[iRight].y - points[iPoint].y);
			dx[1] = points[iPoint].x + (dy - points[iPoint].y) * dax[1];
		}
		
		if(iLeft > iRight)
			break;
		
		int ilx = static_cast<int>(dx[0] + 0.5);
		int irx = static_cast<int>(dx[1] + 0.5);
		
		// Decrease the occupancy probability
		// Simple method, could be calibrated
		unsigned int *dst = (m_cells + iy*m_width + ilx);
		for(int i=ilx;i<irx;i++) {
			unsigned int val = *dst;
			//if(val > minVal) {
			/*	if(weight > (val & 0xff))
					val &= 0xfffffe00;
				else {
					val -= weight;
					val &= 0xfffffeff;
				}*/
				
				// forcing weight = 1 (performance reasons, since other weights proved unnecessary)
				if(val & 0xff) {
					val--;
					val &= 0xfffffeff;// clear unknown bit
				}
			//}
			//val &= 0xfffffeff;
			*dst++ = val;
		}
		
		iy++;
		
		dx[0] += dax[0];
		dx[1] += dax[1];
	}
}// ProbabilityMap::fillPoly


// Fill poly simple, set every cell in the poly to specified value
void ProbabilityMap::fillPoly(
	const vec2d *points,
	int nPoints,
	unsigned int val
) {
	int iPoint = 0;
	int iLeft;
	int iRight;
	double dx[2], dax[2];
	double dy;
	int iy;
	
	iy = static_cast<int>(points[iPoint].y + 0.5);
	dy = static_cast<double>(iy) + 0.5;
	
	iLeft = 1;
	iRight = nPoints-1;
	
	dax[0] = (points[iLeft].x - points[iPoint].x) / (points[iLeft].y - points[iPoint].y);
	dx[0] = points[iPoint].x + (dy - points[iPoint].y) * dax[0];
	
	dax[1] = (points[iRight].x - points[iPoint].x) / (points[iRight].y - points[iPoint].y);
	dx[1] = points[iPoint].x + (dy - points[iPoint].y) * dax[1];
	
	for(;;) {
		dy = static_cast<double>(iy) + 0.5;
		
		while(points[iLeft].y < dy) {
			iLeft++;
			if(iLeft > iRight)
				break;
			
			iPoint = iLeft-1;
			
			dax[0] = (points[iLeft].x - points[iPoint].x) / (points[iLeft].y - points[iPoint].y);
			dx[0] = points[iPoint].x + (dy - points[iPoint].y) * dax[0];
		}
		
		while(points[iRight].y < dy) {
			iRight--;
			if(iLeft > iRight)
				break;
			
			iPoint = iRight+1;
			
			dax[1] = (points[iRight].x - points[iPoint].x) / (points[iRight].y - points[iPoint].y);
			dx[1] = points[iPoint].x + (dy - points[iPoint].y) * dax[1];
		}
		
		if(iLeft > iRight)
			break;
		
		int ilx = static_cast<int>(dx[0] + 0.5);
		int irx = static_cast<int>(dx[1] + 0.5);
		
		// Set cell values
		unsigned int *dst = (m_cells + iy*m_width + ilx);
		for(int i=ilx;i<irx;i++) {
			unsigned int dval = *dst;
			dval = (dval & 0xfffffe00) | val;
			*dst++ = dval;
		}
		
		iy++;
		
		dx[0] += dax[0];
		dx[1] += dax[1];
	}
}// ProbabilityMap::fillPoly
