// ProbailityMapPath.cpp

// ProbabilityMap path finding member functions

#include "ProbabilityMap.h"



// Find a path to an unknown cell
// Dijkstra graph search
// This could be made to incorporate robot turning etc, since it takes longer
// time for the robot to reach a close cell when it must turn a lot
// Now it's a simple search for the closest
// Occupancy proability is considered for distance, so traveling a completely unoccupied
// area is closer than traveling through uncertain areas
// It also tries to keep away from walls where possible (cell-cost checks neighboring cells)
// High-probability occupancy cells are considered impassable
bool ProbabilityMap::findPathToUnknown(const vec2d &pos, vec2d *path, int *pathLena) {
	resetCosts();
	
	// Start point
	vec2i ipos(vec2dtoi(pos));
	std::set<PathPoint, lpathpoint> openList;
	addPoint(openList, ipos.x, ipos.y, 0, ipos.x, ipos.y);
	
	// Loop
	while(!openList.empty()) {
		// Extract least cost point
		PathPoint p = *openList.begin();
		openList.erase(openList.begin());
		
		// Check for unknown
		unsigned int val = m_cells[p.y*m_width + p.x];
		if(val & 0x100) {
			makePath(p.x, p.y, path, pathLena);
			
			return true;
		}
		
		// Check neighbors
		bool lopen = cellOpen(p.x-1, p.y);
		bool ropen = cellOpen(p.x+1, p.y);
		bool uopen = cellOpen(p.x, p.y+1);
		bool dopen = cellOpen(p.x, p.y-1);
		
		// Add adjacent cells if passable
		if(lopen) {
			addPoint(openList, p.x-1, p.y, p.cost+100, p.x, p.y);
			if(uopen)
				addPoint(openList, p.x-1, p.y+1, p.cost+150, p.x, p.y);
			if(dopen)
				addPoint(openList, p.x-1, p.y-1, p.cost+150, p.x, p.y);
		}
		if(ropen) {
			addPoint(openList, p.x+1, p.y, p.cost+100, p.x, p.y);
			if(uopen)
				addPoint(openList, p.x+1, p.y+1, p.cost+150, p.x, p.y);
			if(dopen)
				addPoint(openList, p.x+1, p.y-1, p.cost+150, p.x, p.y);
		}
		if(uopen)
			addPoint(openList, p.x, p.y+1, p.cost+100, p.x, p.y);
		if(dopen)
			addPoint(openList, p.x, p.y-1, p.cost+100, p.x, p.y);
	}
	
	return false;
}// Probabilitymap::findPathToUnknown


// Find path to specified target
// Should be replaced with A* for efficiency (is standard dijkstra)
bool ProbabilityMap::findPathToTarget(const vec2d &pos, const vec2d &target, vec2d *path, int *pathLena) {
	resetCosts();
	
	vec2i itarget(vec2dtoi(target));
	
	// Start point
	vec2i ipos(vec2dtoi(pos));
	std::set<PathPoint, lpathpoint> openList;
	addPoint(openList, ipos.x, ipos.y, 0, ipos.x, ipos.y);
	
	// Loop
	while(!openList.empty()) {
		// Extract least cost point
		PathPoint p = *openList.begin();
		openList.erase(openList.begin());
		
		// Check for unknown
		unsigned int val = m_cells[p.y*m_width + p.x];
		if((int)p.x == itarget.x && (int)p.y == itarget.y) {
			makePath(p.x, p.y, path, pathLena);
			
			return true;
		}
		if(val & 0x100) {
			continue;
		}
		
		// Check neighbors
		bool lopen = cellOpen(p.x-1, p.y);
		bool ropen = cellOpen(p.x+1, p.y);
		bool uopen = cellOpen(p.x, p.y+1);
		bool dopen = cellOpen(p.x, p.y-1);
		
		// Add adjacent cells if passable
		if(lopen) {
			addPoint(openList, p.x-1, p.y, p.cost+100, p.x, p.y);
			if(uopen)
				addPoint(openList, p.x-1, p.y+1, p.cost+150, p.x, p.y);
			if(dopen)
				addPoint(openList, p.x-1, p.y-1, p.cost+150, p.x, p.y);
		}
		if(ropen) {
			addPoint(openList, p.x+1, p.y, p.cost+100, p.x, p.y);
			if(uopen)
				addPoint(openList, p.x+1, p.y+1, p.cost+150, p.x, p.y);
			if(dopen)
				addPoint(openList, p.x+1, p.y-1, p.cost+150, p.x, p.y);
		}
		if(uopen)
			addPoint(openList, p.x, p.y+1, p.cost+100, p.x, p.y);
		if(dopen)
			addPoint(openList, p.x, p.y-1, p.cost+100, p.x, p.y);
	}
	
	return false;
}// Probabilitymap::findPathToTarget


// Check if line is blocked in the map
bool ProbabilityMap::isBlocked(const vec2d &p0, const vec2d &dir, int steps, int width) {
	vec2d	p = p0;
	vec2d	c = cross(dir);
	
	for(int i=0;i<steps;i++) {
		p += dir;
		
		for(int j=-width;j<=width;j++) {
			vec2i ip(vec2dtoi(p + c * static_cast<double>(j)));
			
			if(!cellOpen(ip.x, ip.y))
				return true;
		}
	}
	
	return false;
}// ProbabilityMap::isBlocked


// Check if cell is open
bool ProbabilityMap::cellOpen(unsigned int x, unsigned int y) {
	return (((m_cells[y*m_width + x] & 0xff)) <= m_maxPassableCost);
}// Probabilitymap::cellOpen


// Add point to list, if cost is less than old cost
void ProbabilityMap::addPoint(std::set<PathPoint, lpathpoint> &list, unsigned int x, unsigned int y, unsigned int cost, unsigned int px, unsigned int py) {
	if(cost != 0) {
		// Calculate cost from neighboring cells
		// to keep the robot further from edges where possible
		const int nsize = 5;
		for(int i=-nsize;i<=nsize;i++) {
			for(int j=-nsize;j<=nsize;j++) {
				int mf = 2*nsize*nsize - i*i - j*j;
				mf /= nsize;
				
				unsigned int c = (m_cells[(y+i)*m_width + (x+j)] & 0xff);
				
				cost += mf * ((c));
			}
		}
	}
	
	// Add
	if(cost < m_costs[y*m_width + x]) {
		m_costs[y*m_width + x] = cost;
		// Store x/y of previous cell in path, so that the cheapest path
		// can be reconstructed (each value in m_prev is 32-bit, with 16 bit x and 16 bit y)
		m_prev[y*m_width + x] = ((py & 0xffff) << 16) | (px & 0xffff);
		
		// Avoid duplicates
		for(std::set<PathPoint, lpathpoint>::iterator i=list.begin();i!=list.end();++i) {
			const PathPoint &p(*i);
			
			if(p.x == x && p.y == y) {
				list.erase(i);
				break;
			}
		}
		list.insert(PathPoint(x, y, cost));
	}
}// ProbabilityMap::addPoint


// Make movement path from a detected target point, trace back to starting point
void ProbabilityMap::makePath(unsigned int x, unsigned int y, vec2d *path, int *pathLena) {
	int pathLen = 0;
	
	m_path[pathLen] = vec2i(x, y);
	path[pathLen++] = vec2d(static_cast<double>(x), static_cast<double>(y)) - mapCenter();
	
	while(m_costs[y*m_width + x] != 0) {
		unsigned int prev = m_prev[y*m_width + x];
		x = (prev & 0xffff);
		y = (prev >> 16) & 0xffff;
		
		m_path[pathLen] = vec2i(x, y);
		path[pathLen++] = vec2d(static_cast<double>(x), static_cast<double>(y)) - mapCenter();
	}
	
	m_pathLen = pathLen;
	*pathLena = pathLen;
}// ProbabilityMap::makePath


// Reset cell costs (needed before path finding)
void ProbabilityMap::resetCosts() {
	for(int i=0;i<m_height*m_width;i++)
		m_costs[i] = 0x7fffffff;
}// ProbabilityMap::resetCosts


// Get path data (for viewing)
int ProbabilityMap::getPath(int *path) {
	for(int i=0;i<m_pathLen;i++) {
		const vec2i &p(m_path[i]);
		
		path[i*2] = p.x;
		path[i*2+1] = p.y;
	}
	
	return m_pathLen;
}// ProbabilityMap::getPath


// Discard saved path
void ProbabilityMap::discardPath() {
	m_pathLen = 0;
}// Probabilitymap::discardPath
