// RobotAI.cpp

#include "Aria.h"

#include "HTTPServer.h"
#include "HTTPConnection.h"

#include "ProbabilityMap.h"
#include "LocalizationMap.h"
#include "RobotHandler.h"

#include "ImageProvider.h"


// Include all files here so we can build as a single file
// instead of using a makefile
// (Should have a make-file but it was easier like this when
// moving between VC++ projects and command-line compiling)
/*
#include "CamImage.cpp"
#include "EdgeFinder.cpp"
#include "HTTPConnection.cpp"
#include "HTTPServer.cpp"
#include "ImageProvider.cpp"
#include "JPEGLoader.cpp"
#include "LocalizationMap.cpp"
#include "MyActionLimit.cpp"
#include "ProbabilityMap.cpp"
#include "ProbabilityMapFill.cpp"
#include "ProbabilityMapImage.cpp"
#include "ProbabilityMapPath.cpp"
#include "RobotHandler.cpp"
#include "strstuff.cpp"
*/


bool		g_wantShutdown;

// Main
int main(int argc, char **argv) {
	Aria::init();
	ArThread::init();
	
	g_wantShutdown = false;
	
	// Connect to robot
	ArRobot robot;
	ArSonarDevice sonar;
	
	robot.addRangeDevice(&sonar);
	
	ArSimpleConnector connector(&argc, argv);
	if (!Aria::parseArgs()) {
		Aria::logOptions();
		Aria::shutdown();
		return 1;
	}
	if (!connector.connectRobot(&robot)) {
		ArLog::log(ArLog::Normal, "Could not connect to robot... exiting");
		return 2;
	}
	
	// True parameter means that if the connection is lost, then the 
	// run loop ends.
	robot.runAsync(true);
	
	// The probability map
	int mapWidth = 2048;
	int mapHeight = 2048;
	ProbabilityMap pmap(mapWidth, mapHeight);
	
	// Projection map
	LocalizationMap prmap;
	
	// Image provider
	ImageProvider imgProvider;
	
	imgProvider.runAsync();
	
	// Robot handler
	RobotHandler handler(&robot, &pmap, &imgProvider, &prmap);
	
	handler.init();
	
	// Run HTTP server, for handling HTTP connections with orders or requests
	HTTPConnection httpHandler(&pmap, &handler, &imgProvider);
	
	HTTPServer httpServer(&pmap, &handler, &imgProvider, &httpHandler);
	
	httpServer.runAsync();
	
	// Main loop
	bool loop = true;
	unsigned int oldTime = ArUtil::getTime();
	unsigned int netTime = oldTime;
	while(loop) {
		//ArUtil::sleep(40);
		unsigned int newTime = ArUtil::getTime();
		if(newTime < (oldTime+40)) {
			ArUtil::sleep((oldTime+40)-newTime);
		}
		oldTime = newTime;
		
		// Handle robot
		handler.updateMapWithRobotData();
		handler.handleRobotBehavior();
		
		// Handle network (if enough time past since last)
		if(newTime > netTime+150) {
			httpHandler.handleNetwork();
			
			netTime = newTime;
		}
		
		// Check for shutdown
		if(g_wantShutdown)
			loop = false;
	}
	
	// Destroy robot handler
	handler.destroy();
	
	// Stop HTTP server
	httpServer.stopRunning();
	httpServer.myStop();
	do {
		ArUtil::sleep(100);
	} while(httpServer.getRunningWithLock());
	
	// Stop image provider
	imgProvider.stopRunning();
	imgProvider.myStop();
	do {
		ArUtil::sleep(100);
	} while(imgProvider.getRunningWithLock());
	
	// Stop robot
	robot.stopRunning();
	
	
	Aria::uninit();
	
	return 0;
}


void doShutdown() {
	g_wantShutdown = true;
}
