// RobotHandler.cpp

// This class handles the robot behavior

// It is too big and should be refactored...

// It processes input and passes it along to the ProbabilityMap and the LocalizationMap
// It also receives orders from admin-connections

// When moving the robot this class asks the ProbabilityMap for movement path through
// the map and activates the appropriate actions to move the robot with ARIA


#include "RobotHandler.h"
#include "CamImage.h"
#include "EdgeFinder.h"
#include "ImageProvider.h"
#include "ProbabilityMap.h"
#include "LocalizationMap.h"

#include <cmath>


// Construct
RobotHandler::RobotHandler(
	ArRobot *robot,
	ProbabilityMap *map,
	ImageProvider *imgProvider,
	LocalizationMap *prmap
) :
	// Robot and map
	m_robot(robot),
	m_map(map),
	m_mapScale(1.0/50.0),
	m_LocalizationMap(prmap),
	m_imgProvider(imgProvider),
	
	// Finit state machine
	m_hasStop(false),
	
	m_wanderMode(false),
	
	m_gotoMode(false),
	
	m_state(rsIdle),
	m_orders(roStop),
	//m_orders(roExplore),
	//m_orders(roWander),
	
	m_path(0),
	m_pathLen(0),
	m_curPathPoint(0),
	m_robotHeading(0.0),
	m_timer(0),
	m_hasTarget(false),
	
	m_hasNewActors(false),
	
	// Settings
	m_sonarEnabled(true),
	m_camEnabled(false),
	m_updateMap(true),
	
	m_camFrames(0),
	
	// Odometry error correction
	m_correctOdo(true),
	m_thetas(0),
	m_poses(0),
	m_nThetas(5),
	
	m_hasOldPose(false)
{
	// Path-finding path storage
	const int maxPathLen = 1000;
	
	m_path = new vec2d[maxPathLen];
	
	// Edge detection settings
	m_settings.thres = 45;
	
	m_settings.minEdgeLen = 25;
	m_settings.minStartLen = 15;
	m_settings.maxYMiss = 15;
	m_settings.minCenterMoveY = 4;
	
	m_settings.minXDist = 15;
	m_settings.maxXDist = 100;
	m_settings.dirNum = 2;
	m_settings.dirDen = 3;
	
	m_thetas = new double[m_nThetas];
	m_poses = new ArPose[m_nThetas];
	for(int i=0;i<m_nThetas;i++)
		m_thetas[i] = 0.0;
}// RobotHandler::RobotHandler


// Destruct
RobotHandler::~RobotHandler() {
	delete[] m_path;
	
	delete[] m_thetas;
	delete[] m_poses;
}// RobotHandler::~RobotHandler


// Init
void RobotHandler::init() {
	// Max rotation speed
	m_robot->setRotVelMax(45.0);
	
	m_oldPose = m_robot->getPose();
	m_hasOldPose = true;
	
	// Wander
	m_wander = new ArActionGroup(m_robot);
	m_wander->addAction(new ArActionStallRecover, 100);
	m_wander->addAction(new MyActionLimit(m_map, this, m_mapScale), 70);
	m_wander->addAction(new ArActionAvoidFront("", 400.0, 0.0, 30), 75);
	m_wander->addAction(new ArActionAvoidSide(), 60);
	m_wander->addAction(new ArActionAvoidFront("", 800.0, 200.0), 50);
	m_wander->addAction(new ArActionConstantVelocity("", 300.0), 25);
	m_wander->deactivate();
	
	// Explore
	// Store goto action pointer to be able to change the target
	m_actionGoto = new ArActionGoto("goto", ArPose(0.0, 0.0, 0.0), 200.0, 200.0, 75.0, 2.0);
	m_actionGoto->cancelGoal();
	m_goto = new ArActionGroup(m_robot);
	m_goto->addAction(new ArActionStallRecover, 100);
	m_goto->addAction(new MyActionLimit(m_map, this, m_mapScale), 70);
	m_goto->addAction(m_actionGoto, 40);
	m_goto->deactivate();
	
	// Action to make the robot stop when without orders
	m_actionStop = new ArActionStop("stop");
	
	m_robot->addAction(m_actionStop, 10);
	m_hasStop = true;
	
	// Enable robot motors
	ArUtil::sleep(100);
	m_robot->enableMotors();
}// RobotHandler::init


// Destroy
void RobotHandler::destroy() {
	m_robot->remAction(m_actionStop);
	delete m_actionStop;
	
	//m_wander->deleteActions();
	delete m_wander;
	
	delete m_goto;
	
	delete m_actionGoto;
}// RobotHandler::destroy


// Update map an handle robot data
// Actually this method handles everything except the actual movement
// It does odometry error correction and checks camera image for dynamic
// obstacles in addition to updating the map with sonar readings
void RobotHandler::updateMapWithRobotData() {
	m_robotPos = m_robot->getPose();
	m_robotHeading = m_robotPos.getThRad();
	
	Ray rays[8];
	Ray prays[8];
	
	// Get sonar readings
	if(m_sonarEnabled) {
		for(int i=0;i<8;i++) {
			ArSensorReading *sr = m_robot->getSonarReading(i);
			
			Ray &r = rays[i];
			Ray &pr = prays[i];
			
			r.valid = false;
			pr.valid = false;
			
			if(!sr->getIgnoreThisReading()) {
				double x = sr->getX();
				double y = sr->getY();
				double sx = sr->getSensorX();
				double sy = sr->getSensorY();
				
				r.p0 = vec2d(sr->getXTaken()+sx, sr->getYTaken()+sy) * m_mapScale + m_map->mapCenter();
				r.p1 = vec2d(x, y) * m_mapScale + m_map->mapCenter();
				r.endHit = sr->getRange() < 5000;
				r.valid = sr->getRange() > 0;
				
				pr.p0 = vec2d(m_robotPos.getX()+sx, m_robotPos.getY()+sy);
				pr.p1 = vec2d(x, y);
				pr.endHit = sr->getRange() < 5000;
				pr.valid = sr->getRange() > 0;
			}
		}
	}
	
	// Input sonar readings to map
	if(m_sonarEnabled && m_updateMap) {
		m_map->input(vec2d(m_robotPos.getX(), m_robotPos.getY()) * m_mapScale + m_map->mapCenter(), rays, 8);
	}
	else {
		// Even if sonars are disabled, still add a dirty-rect around the robot (to update map rect)
		vec2i p = vec2dtoi(vec2d(m_robotPos.getX(), m_robotPos.getY()) * m_mapScale + m_map->mapCenter());
		Rect rect;
		
		rect.left = p.x - 20;
		rect.top = p.y - 20;
		rect.width = 40;
		rect.height = 40;
		m_map->addDirtyRect(rect);
	}
	
	m_hasNewActors = false;
	
	// Use cam image
	bool useCam = m_camEnabled;
	if(useCam) {
		// Only use cam image every 8th iteration
		m_camFrames++;
		if(m_camFrames > 8) {
			CamImage *img = m_imgProvider->getImage();
			
			if(img) {
				if(img->decodeImage()) {
					std::vector<Edge> edges;
					
					// Find edges in image
					findEdges(img->pixels, img->w, img->h, edges, &m_settings);
					
					// Update the list of actors from what was detected in the image
					m_actors.clear();
					if(edges.size() > 0) {
						for(size_t i=0;i<edges.size();i++) {
							const Edge &edge(edges[i]);
							vec2d imgp = vec2d(static_cast<double>(edge.ex), static_cast<double>(edge.ey));
							vec2d imgsize = vec2d(static_cast<double>(img->w), static_cast<double>(img->h));
							vec2d imgray;
							
							// Get the robot-space coordinates of the point in the image the actor is at
							if(getImgRay(imgp, imgsize, imgray)) {
								// Transform to world-space
								vec2d ray;
								ray.x = imgray.x * std::cos(m_robotHeading) - imgray.y * std::sin(m_robotHeading);
								ray.y = imgray.x * std::sin(m_robotHeading) + imgray.y * std::cos(m_robotHeading);
								vec2d rpos = vec2d(m_robotPos.getX(), m_robotPos.getY());
								vec2d worldp = rpos + ray;
								
								// Remember the actor to avoid it when moving
								m_actors.push_back(Actor(worldp));
								m_hasNewActors = true;
							}
						}
					}
				}
				
				img->release();
			}
			
			m_camFrames = 0;
		}
		
		prepareForActors();
	}
	else {
		// No cam, clear any old actors
		m_actors.clear();
		prepareForActors();
	}
	
	// Odometry drift correction
	// Uses data from the LocalizationMap to correct drift
	// This should be refactored and updated to allow for better estimates
	if(m_correctOdo) {
		// Input sonar data to the wall detection
		if(m_LocalizationMap->input(vec2d(m_robotPos.getX(), m_robotPos.getY()), prays, 8)) {
			const double pi = 3.14159265358979;
			
			// Check if there is a wall to adjust orientation to
			double t = m_LocalizationMap->getTheta();
			if(t != 0.0) {
				// Check so that there is enough evidence for a likely drift
				// (enough walls leaning in the same direction)
				double mind = pi * 0.2;
				double d;
				double rd;
				
				d = t;
				rd = d;
				d = t-pi;
				if(std::abs(d) < std::abs(rd))
					rd = d;
				d = t-(pi*0.5);
				if(std::abs(d) < std::abs(rd))
					rd = d;
				
				for(int i=0;i<m_nThetas-1;i++) {
					m_thetas[i] = m_thetas[i+1];
					m_poses[i] = m_poses[i+1];
				}
				m_thetas[m_nThetas-1] = rd;
				m_poses[m_nThetas-1] = m_robot->getPose();
				
				int bad = 5;
				if(m_thetas[0] > 0.0) {
					bad = 0;
					for(int i=0;i<m_nThetas;i++) {
						if(m_thetas[i] <= 0.0) {
							bad++;
						}
					}
				}
				else if(m_thetas[0] < 0.0) {
					bad = 0;
					for(int i=0;i<m_nThetas;i++) {
						if(m_thetas[i] >= 0.0) {
							bad++;
						}
					}
				}
				
				// If we found walls to correct to
				if(bad == 0) {
					// Get the average error of the last detected walls
					double tt = 0.0;
					int nt = 0;
					for(int i=0;i<m_nThetas;i++) {
						if((m_thetas[i] < 0.0 && m_thetas[0] < 0.0) || (m_thetas[i] > 0.0 && m_thetas[0] > 0.0)) {
							tt += m_thetas[i];
							nt++;
						}
					}
					tt /= static_cast<double>(nt);
					
					// Only correct if the error isn't too big
					// (too big error probably means false walls)
					// (But if the detection takes too long it could instead miss walls, is a calibrated estimate)
					if(std::abs(tt) < mind) {
						ArPose p = m_robot->getPose();
						
						p.setThRad(p.getThRad()-tt);
						
						if(m_hasOldPose) {
							// Calculate new position from rotation drift
							ArPose op = m_oldPose;
							vec2d vt = vec2d(0.0, 0.0);
							for(int i=0;i<m_nThetas;i++) {
								if((m_thetas[i] < 0.0 && m_thetas[0] < 0.0) || (m_thetas[i] > 0.0 && m_thetas[0] > 0.0)) {
									vec2d v = vec2d(m_poses[i].getX()-op.getX(), m_poses[i].getY()-op.getY());
									
									v.rotate(tt/static_cast<double>(nt));
									vt += v;
									
									op = m_poses[i];
								}
							}
							
							p.setX(m_oldPose.getX()+vt.x);
							p.setY(m_oldPose.getY()+vt.y);
						}
						
						m_robot->moveTo(p);
						
						for(int i=0;i<m_nThetas;i++)
							m_thetas[i] = 0.0;
						
						m_oldPose = p;
						m_hasOldPose = true;
					}
				}
			}
		}
	}
}// RobotHandler::updateMapWithRobotData


// Handle robot behavior
// The robot is handled as a finite state machine
// This could be written nicer...
void RobotHandler::handleRobotBehavior() {
	m_timer++;
	
	vec2d target;
	bool hasTarget = false;
	
	// Handle orders
	if(m_hasTarget) {
		m_hasTarget = false;
		hasTarget = true;
		Rect mapRect = m_map->getTotalRect();
		target.x = (double)mapRect.left + m_target.x;
		target.y = (double)mapRect.top + m_target.y;
	}
	
	if(m_orders != roExplore && m_orders != roGoto) {
		m_map->discardPath();
	}
	
	// Check for need to deactivate wander mode
	if(m_wanderMode) {
		if(m_orders != roWander || hasTarget) {
			m_wander->deactivate();
			
			m_actionStop->activate();
			m_hasStop = true;
			
			m_wanderMode = false;
		}
		else {
			return;
		}
	}
	
	// Check for need to deactivate goto mode
	if(m_gotoMode) {
		if(m_orders != roExplore && m_orders != roGoto) {
			m_goto->deactivate();
			
			m_actionStop->activate();
			m_hasStop = true;
			
			m_gotoMode = false;
		}
	}
	
	// Wander mode
	if(m_orders == roWander && !m_wanderMode && !hasTarget) {
		m_wanderMode = true;
		m_wander->activateExclusive();
		return;
	}
	
	// Goto target
	if(hasTarget) {
		m_orders = roGoto;
		
		// Ask the map for a path to the target
		if(m_map->findPathToTarget(vec2d(m_robotPos.getX(), m_robotPos.getY()) * m_mapScale + m_map->mapCenter(), target, m_path, &m_pathLen)) {
			printf("found target\n");
			m_state = rsGoto;
			m_timer = 1;
			m_curPathPoint = m_pathLen-1;
			m_targetPoint = target;
			
			if(m_hasStop) {
				//m_robot->remAction(m_actionStop);
				m_actionStop->deactivate();
				m_hasStop = false;
			}
		}
		else
			printf("fail\n");
	}
	
	// Goto mode
	if((m_orders == roExplore || m_orders == roGoto) && !m_gotoMode) {
		m_gotoMode = true;
		m_goto->activateExclusive();
	}
	
	// Handle state
	switch(m_state) {
		case rsIdle:
			handleIdleState();
		break;
		
		case rsExplorePath:
			handleExplorePathState();
		break;
		
		case rsGoto:
			handleGoto();
		break;
	}
}// RobotHandler::handleRobotBehavior


// Prepare for dynamic actors
void RobotHandler::prepareForActors() {
	if(m_actors.size() > 0) {
		// This used to do something..
	}
}// RobotHandler::prepareForActors


// Get actor vector
std::vector<Actor>& RobotHandler::getActors() {
	return m_actors;
}// RobotHandler::getActors


// Handle idle state
void RobotHandler::handleIdleState() {
	if(m_orders == roStop) {
		if(!m_hasStop) {
			m_actionStop->activate();
			m_hasStop = true;
		}
	}
	else if(m_orders == roExplore) {
		// Find new path and set state to follow it
		if(m_map->findPathToUnknown(vec2d(m_robotPos.getX(), m_robotPos.getY()) * m_mapScale + m_map->mapCenter(), m_path, &m_pathLen)) {
			m_state = rsExplorePath;
			m_timer = 0;
			m_curPathPoint = m_pathLen-1;
			
			if(m_hasStop) {
				m_actionStop->deactivate();
				m_hasStop = false;
			}
		}
		else {
			m_actionGoto->setGoal(m_robotPos);
			
			if(!m_hasStop) {
				m_actionStop->activate();
				m_hasStop = true;
			}
			
			m_orders = roStop;// Stop if there is nowhere to explore
		}
	}
	else {
		m_orders = roStop;
	}
}// RobotHandler::handleIdleState


// Handle explore path state
void RobotHandler::handleExplorePathState() {
	// Follow path
	if(m_timer == 1 || m_actionGoto->haveAchievedGoal()) {
		if(m_curPathPoint > 10)
			m_curPathPoint -= 10;
		else
			m_curPathPoint = 0;
		
		vec2d pos = m_path[m_curPathPoint] / m_mapScale;
		ArPose pose = ArPose(pos.x, pos.y, 0);
		
		m_actionGoto->setGoal(pose);
	}
	
	// Go to idle, and a recalculated path (possibly to a new target)
	// will be found next iteration
	if(m_timer > 15) {
		m_state = rsIdle;
		m_timer = 0;
	}
}// RobotHandler::handleExplorePathState


// Handle goto state
void RobotHandler::handleGoto() {
	// Follow path
	if(m_timer == 1 || m_actionGoto->haveAchievedGoal()) {
		if(m_curPathPoint < 5) {
			m_actionGoto->cancelGoal();
			m_state = rsIdle;
			m_timer = 0;
			return;
		}
		
		if(m_curPathPoint > 10)
			m_curPathPoint -= 10;
		else
			m_curPathPoint = 0;
		
		vec2d pos = m_path[m_curPathPoint] / m_mapScale;
		ArPose pose = ArPose(pos.x, pos.y, 0);
		
		m_actionGoto->setGoal(pose);
	}
	
	// Recalculate path regularly (map and dynamic obstacles can be changed)
	if(m_timer > 20) {
		if(m_map->findPathToTarget(vec2d(m_robotPos.getX(), m_robotPos.getY()) * m_mapScale + m_map->mapCenter(), m_targetPoint, m_path, &m_pathLen)) {
			m_state = rsGoto;
			m_timer = 0;
			m_curPathPoint = m_pathLen-1;
			
			if(m_hasStop) {
				m_actionStop->deactivate();
				m_hasStop = false;
			}
		}
	}
}// RobotHandler::handleGoto


// Get robot x
int RobotHandler::getRobotX() const {
	return static_cast<int>((vec2d(m_robotPos.getX(), m_robotPos.getY()) * m_mapScale + m_map->mapCenter()).x);
}// RobotHandler::getRobotX
// Get robot y
int RobotHandler::getRobotY() const {
	return static_cast<int>((vec2d(m_robotPos.getX(), m_robotPos.getY()) * m_mapScale + m_map->mapCenter()).y);
}// RobotHandler::getRobotY
// Get robot heading
int RobotHandler::getRobotHeading() const {
	return static_cast<int>(m_robotHeading * 1000.0);
}// RobotHandler::getRobotHeading
// Get robot pose
ArPose RobotHandler::getRobotPose() {
	return m_robotPos;
}// RobotHandler::getRobotPose
// Get robot front
int RobotHandler::getRobotFront() {
	return static_cast<int>(m_robot->getRobotLengthFront() * m_mapScale);
}// RobotHandler::getRobotFront
// Get robot rear
int RobotHandler::getRobotRear() {
	return static_cast<int>(m_robot->getRobotLengthRear() * m_mapScale);
}// RobotHandler::getRobotRear
// Get robot width
int RobotHandler::getRobotWidth() {
	return static_cast<int>(m_robot->getRobotWidth() * m_mapScale);
}// RobotHandler::getRobotWidth


// Order goto map position
void RobotHandler::gotoMapPos(int x, int y) {
	m_hasTarget = true;
	m_target = vec2d(static_cast<double>(x), static_cast<double>(y));
}// RobotHandler::gotoMapPos


// Orders
void RobotHandler::inputOrders(int iorders) {
	RobotOrders orders = roNone;
	
	switch(iorders) {
		case 0:
		break;
		case 1:
			orders = roExplore;
		break;
		case 3:
			orders = roWander;
		break;
		case 4:
			orders = roStop;
		break;
	}
	
	if(orders != roNone) {
		m_orders = orders;
		m_state = rsIdle;
	}
}// RobotHandler::inputOrders


// Enable cam
void RobotHandler::enableCam(bool enable) {
	m_camEnabled = enable;
}// RobotHandler::enableCam


// Enable sonar
void RobotHandler::enableSonar(bool enable) {
	if(enable && !m_sonarEnabled) {
		m_robot->enableSonar();
	}
	else if(!enable && m_sonarEnabled) {
		m_robot->disableSonar();
	}
	m_sonarEnabled = enable;
}// RobotHandler::enableSonar


// Set map updates
void RobotHandler::setUpdateMap(bool enable) {
	m_updateMap = enable;
}// RobotHandler::setUpdateMap


// Set correct odometry
void RobotHandler::setCorrectOdo(bool enable) {
	m_correctOdo = enable;
	m_hasOldPose = false;
	for(int i=0;i<m_nThetas;i++)
		m_thetas[i] = 0.0;
}// RobotHandler::setCorrectOdo


// Get ray from image position
// Bilinear interpolation over a matrix of known world-space rays for a grid on the image
// This method doesn't really belong in RobotHandler.. this should be in some camera class...
bool RobotHandler::getImgRay(const vec2d &pix, const vec2d &img, vec2d &ray) {
	// This matrix is manually measured and must be calibrated when camera position is changed
	static const int th = 8;
	static const int tw = 8;
	static const vec2d table[64] = {
		vec2d(-0,0),vec2d(-0,0),vec2d(-0,0),vec2d(-0,0),vec2d(-0,0),vec2d(-0,0),vec2d(-0,0),vec2d(-0,0),
		vec2d(-0,0),vec2d(-0,0),vec2d(-0,0),vec2d(-0,0),vec2d(-0,0),vec2d(-0,0),vec2d(-0,0),vec2d(-0,0),
		vec2d(-1239,1889),vec2d(-760,1619),vec2d(-436,1651),vec2d(-118,1551),vec2d(104,1555),vec2d(364,1643),vec2d(754,1726),vec2d(1182,1851),
		vec2d(-718,1128),vec2d(-422,1064),vec2d(-230,1049),vec2d(-84,970),vec2d(72,1022),vec2d(213,1064),vec2d(389,1076),vec2d(743,1166),
		vec2d(-478,830),vec2d(-397,903),vec2d(-207,867),vec2d(-76,903),vec2d(63,893),vec2d(203,891),vec2d(302,902),vec2d(454,836),
		vec2d(-364,654),vec2d(-245,642),vec2d(-178,671),vec2d(-55,660),vec2d(48,658),vec2d(177,672),vec2d(237,643),vec2d(365,657),
		vec2d(-313,549),vec2d(-212,563),vec2d(-124,553),vec2d(-48,563),vec2d(39,565),vec2d(119,555),vec2d(209,568),vec2d(311,551),
		vec2d(-300,430),vec2d(-200,450),vec2d(-115,460),vec2d(-35,470),vec2d(35,470),vec2d(115,460),vec2d(200,450),vec2d(300,430)
	};
	
	// Linear interpolation to estimate position
	vec2d b0, b1, b2, b3;
	
	vec2d p;
	
	p.x = (pix.x * static_cast<double>(tw)) / img.x;
	p.y = (pix.y * static_cast<double>(th)) / img.y;
	
	vec2i ip = vec2dtoi(p);
	if(ip.x < 0) ip.x = 0;
	if(ip.y < 0) ip.y = 0;
	if(ip.x > static_cast<int>(img.x)-2) ip.x = static_cast<int>(img.x)-2;
	if(ip.y > static_cast<int>(img.y)-2) ip.y = static_cast<int>(img.y)-2;
	
	p.x -= std::floor(p.x);
	p.y -= std::floor(p.y);
	
	b0 = table[ip.y*tw+ip.x];
	b1 = table[ip.y*tw+ip.x+1];
	b2 = table[(ip.y+1)*tw+ip.x];
	vec2d t11 = table[(ip.y+1)*tw+ip.x+1];
	
	if(b0.y == 0.0 || b1.y == 0.0 || b2.y == 0.0 || t11.y == 0.0) return false;
	
	b0.y = 1.0 / b0.y;
	b1.y = 1.0 / b1.y;
	b2.y = 1.0 / b2.y;
	t11.y = 1.0 / t11.y;
	
	b3 = b0 - b1 - b2 + t11;
	b1 -= b0;
	b2 -= b0;
	
	vec2d tray = b0 + b1 * p.x + b2 * p.y + b3 * p.x * p.y;
	
	tray.y = 1.0 / tray.y;
	
	ray.x = tray.y;
	ray.y = -tray.x;
	
	if(ray.length() > 250.0)
		return true;
	else
		return false;
}// RobotHandler::getImgRay


// Get robot status
RobotStatus RobotHandler::getStatus() {
	RobotStatus status;
	
	status.sonar = m_sonarEnabled;
	status.camera = m_camEnabled;
	status.orders = 0;
	switch(m_orders) {
		case roExplore: status.orders = 1; break;
		case roGoto: status.orders = 2; break;
		case roWander: status.orders = 3; break;
		case roStop: status.orders = 4; break;
	}
	status.action = 0;
	switch(m_state) {
		case rsExplorePath: status.action = 1; break;
		case rsGoto: status.action = 2; break;
	}
	status.map = m_updateMap;
	status.odo = m_correctOdo;
	
	return status;
}// RobotHandler::getStatus


// Get sonar status string
const char* RobotHandler::getSonarStatus() {
	const char *sonarEnabled = "Enabled";
	const char *sonarDisabled = "Disabled";
	
	const char *status = sonarDisabled;
	
	if(m_sonarEnabled)
		status = sonarEnabled;
	
	return status;
}// RobotHandler::getSonarStatus


// Get camera status string
const char* RobotHandler::getCamStatus() {
	const char *camEnabled = "Enabled";
	const char *camDisabled = "Disabled";
	
	const char *status = camDisabled;
	
	if(m_camEnabled)
		status = camEnabled;
	
	return status;
}// RobotHandler::getCamStatus


// Get robot orders string
const char* RobotHandler::getOrders() {
	const char *ordersNone = "None";
	const char *ordersExplore = "Explore";
	const char *ordersGoto = "Travel to Target";
	const char *ordersWander = "Wander Randomly";
	const char *ordersStop = "Stop";
	
	const char *status = ordersNone;
	
	switch(m_orders) {
		case roNone:
		break;
		case roExplore:
			status = ordersExplore;
		break;
		case roGoto:
			status = ordersGoto;
		break;
		case roWander:
			status = ordersWander;
		break;
		case roStop:
			status = ordersStop;
		break;
	}
	
	return status;
}// RobotHandler::getOrders


// Get robot action string
const char* RobotHandler::getAction() {
	const char *actionIdle = "Idle";
	const char *actionExplore = "Explore path";
	const char *actionGoto = "Moving to target";
	
	const char *status = actionIdle;
	
	switch(m_state) {
		case rsIdle:
		break;
		case rsExplorePath:
			status = actionExplore;
		break;
		case rsGoto:
			status = actionGoto;
		break;
	}
	
	return status;
}// RobotHandler::getAction


// Set edge detection settings
void RobotHandler::setEdgeSettings(const EdgeSettings &settings) {
	m_settings = settings;
}// RobotHandler::setEdgeSettings
