// RobotHandler.h

#ifndef My__RobotHandler
#define My__RobotHandler


// Dependencies

#include "Aria.h"
#include "MyActionLimit.h"
#include "vec2.h"
#include "EdgeFinder.h"

class ImageProvider;
class ProbabilityMap;
class LocalizationMap;


// Robot state

enum RobotState {
	rsIdle,
	rsExplorePath,
	rsGoto
};


// Robot orders

enum RobotOrders {
	roNone,
	roExplore,
	roGoto,
	roWander,
	roStop
};


// RobotStatus struct

struct RobotStatus {
	bool		sonar;
	bool		camera;
	int			orders;
	int			action;
	bool		map;
	bool		odo;
};


// Actor

struct Actor {
	vec2d		pos;
	
	inline Actor(const vec2d &p) : pos(p) {}
};


// RobotHandler class

class RobotHandler {
	public:
		RobotHandler(
			ArRobot *robot,
			ProbabilityMap *map,
			ImageProvider *imgProvider,
			LocalizationMap *prmap
		);
		~RobotHandler();
		
		void init();
		void destroy();
		
		void updateMapWithRobotData();
		void prepareForActors();
		std::vector<Actor>& getActors();
		
		void handleRobotBehavior();
		
		int getRobotX() const;
		int getRobotY() const;
		int getRobotHeading() const;
		ArPose getRobotPose();
		int getRobotFront();
		int getRobotRear();
		int getRobotWidth();
		
		void gotoMapPos(int x, int y);
		void inputOrders(int iorders);
		void enableCam(bool enable);
		void enableSonar(bool enable);
		void setUpdateMap(bool enable);
		void setCorrectOdo(bool enable);
		
		RobotStatus getStatus();
		const char* getSonarStatus();
		const char* getCamStatus();
		const char* getOrders();
		const char* getAction();
		
		void setEdgeSettings(const EdgeSettings &settings);
	
	private:
		void handleIdleState();
		void handleExplorePathState();
		void handleGoto();
		
		bool getImgRay(const vec2d &pix, const vec2d &img, vec2d &ray);
	
	private:
		ArRobot					*m_robot;
		ProbabilityMap			*m_map;
		double					m_mapScale;
		LocalizationMap			*m_LocalizationMap;
		ImageProvider			*m_imgProvider;
		
		ArActionGoto			*m_actionGoto;
		ArActionStop			*m_actionStop;
		bool					m_hasStop;
		
		ArActionGroup			*m_wander;
		bool					m_wanderMode;
		
		ArActionGroup			*m_goto;
		bool					m_gotoMode;
		
		RobotState				m_state;
		RobotOrders				m_orders;
		
		ArPose					m_robotPos;
		double					m_robotHeading;
		
		vec2d					*m_path;
		int						m_pathLen;
		int						m_curPathPoint;
		
		int						m_timer;
		
		bool					m_hasTarget;
		vec2d					m_target;
		
		vec2d					m_targetPoint;
		
		std::vector<Actor>		m_actors;
		bool					m_hasNewActors;
		EdgeSettings			m_settings;
		
		bool					m_sonarEnabled;
		bool					m_camEnabled;
		bool					m_updateMap;
		
		int						m_camFrames;
		
		bool					m_correctOdo;
		double					*m_thetas;
		ArPose					*m_poses;
		int						m_nThetas;
		
		ArPose					m_oldPose;
		bool					m_hasOldPose;
};


#endif
