// vec2.h

#ifndef My__vec2
#define My__vec2


// Dependencies
#include <cmath>


// vec2 class
// 2d-vector
template <typename T>
class vec2 {
	public:
		T		x, y;
		
		inline vec2() : x(0), y(0) {}
		inline vec2(T xx, T yy) : x(xx), y(yy) {}
		inline vec2(const vec2 &source) : x(source.x), y(source.y) {}
		
		inline vec2& operator=(const vec2 &v) {
			x = v.x; y = v.y; return *this;
		}
		inline vec2 operator-() const {
			return vec2(-x, -y);
		}
		
		inline vec2 operator+(const vec2 &v) const {
			return vec2(x+v.x, y+v.y);
		}
		inline vec2 operator-(const vec2 &v) const {
			return vec2(x-v.x, y-v.y);
		}
		inline vec2 operator*(const vec2 &v) const {
			return vec2(x*v.x, y*v.y);
		}
		inline vec2 operator/(const vec2 &v) const {
			return vec2(x/v.x, y/v.y);
		}
		inline vec2 operator*(T d) const {
			return vec2(x*d, y*d);
		}
		inline vec2 operator/(T d) const {
			return vec2(x/d, y/d);
		}
		
		inline vec2& operator+=(const vec2 &v) {
			x += v.x; y += v.y; return *this;
		}
		inline vec2& operator-=(const vec2 &v) {
			x -= v.x; y -= v.y; return *this;
		}
		inline vec2& operator*=(const vec2 &v) {
			x *= v.x; y *= v.y; return *this;
		}
		inline vec2& operator/=(const vec2 &v) {
			x /= v.x; y /= v.y; return *this;
		}
		inline vec2& operator*=(T d) {
			x *= d; y *= d; return *this;
		}
		inline vec2& operator/=(T d) {
			x /= d; y /= d; return *this;
		}
		
		inline T length2() const {
			return x*x + y*y;
		}
		inline T length() const {
			return static_cast<T>(std::sqrt(static_cast<double>(length2())));
		}
		
		inline void rotate(double a) {
			T nx = static_cast<T>(static_cast<double>(x) * std::cos(a) - static_cast<double>(y) * std::sin(a));
			T ny = static_cast<T>(static_cast<double>(y) * std::cos(a) + static_cast<double>(x) * std::sin(a));
			
			x = nx;
			y = ny;
		}
		
		inline vec2& normalize() {
			T l = length();
			if(l < 0.01) l = 0.01;
			x /= l; y /= l; return *this;
		}
		inline vec2 unit() const {
			T l = length();
			if(l < 0.01) l = 0.01;
			return vec2(x/l, y/l);
		}
};

template <typename T>
inline vec2<T> operator*(T d, const vec2<T> &v) {
	return v * d;
}

template <typename T>
inline T dot(const vec2<T> &v0, const vec2<T> &v1) {
	return (v0.x * v0.x + v0.y * v0.y);
}

template <typename T>
inline vec2<T> cross(const vec2<T> &v) {
	return vec2<T>(v.y, -v.x);
}

template <typename T>
inline T dist(const vec2<T> &p0, const vec2<T> &p1) {
	return (p1 - p0).length();
}

template <typename T>
inline T dist2(const vec2<T> &p0, const vec2<T> &p1) {
	return (p1 - p0).length2();
}


// Vector types
typedef vec2<int> vec2i;
typedef vec2<float> vec2f;
typedef vec2<double> vec2d;


// Line
class line2d {
	public:
		vec2d	p0, p1;
		
		inline line2d() {}
		inline line2d(const vec2d &pp0, const vec2d &pp1) : p0(pp0), p1(pp1) {}
		inline line2d(const line2d &source) : p0(source.p0), p1(source.p1) {}
		
		inline line2d operator*(double d) const {
			return line2d(p0*d, p1*d);
		}
		
		inline double length() const {
			return (p1 - p0).length();
		}
		
		inline vec2d normal() const {
			return cross(p1-p0).normalize();
		}
};

// Line/vector distance
inline double dist(const line2d &l, const vec2d &p) {
	vec2d n = l.normal();
	double ld = dot(l.p0, n);
	
	double d = dot(p, n) - ld;
	
	return d;
}

// Cut a line by another line, output intersection point and lengths of the segment on each side
inline bool cut(const line2d &l0, const line2d &l1, vec2d &ip, double &len0, double &len1) {
	double d = (l1.p1.y - l1.p0.y) * (l0.p1.x - l0.p0.x) - (l1.p1.x - l1.p0.x) * (l0.p1.y - l0.p0.y);
	
	if(d*d > 0.0001) {
		double ua = ((l1.p1.x - l1.p0.x) * (l0.p0.y - l1.p0.y) - (l1.p1.y - l1.p0.y) * (l0.p0.x - l1.p0.x)) / d;
		
		if(ua > 0.0 && ua < 1.0) {
			double ub = ((l0.p1.x - l0.p0.x) * (l0.p0.y - l1.p0.y) - (l0.p1.y - l0.p0.y) * (l0.p0.x - l1.p0.x)) / d;
			
			if(ub > 0.0 && ub < 1.0) {
				ip = l0.p0 + ua * (l0.p1 - l0.p0);
				
				len0 = ua * l0.length();
				len1 = (1.0 - ua) * l0.length();
				
				return true;
			}
		}
	}
	
	return false;
}

// Get integer vector from floor of double vector
inline vec2i vec2dtoi(const vec2d &v) {
	return vec2i(static_cast<int>(v.x), static_cast<int>(v.y));
}


#endif
